from __future__ import annotations

import math
import stat
from typing import TYPE_CHECKING, BinaryIO

from dissect.ntfs import NTFS, NTFS_SIGNATURE, IndexEntry, MftRecord
from dissect.ntfs.exceptions import Error as NtfsError
from dissect.ntfs.exceptions import FileNotFoundError as NtfsFileNotFoundError
from dissect.ntfs.exceptions import NotADirectoryError as NtfsNotADirectoryError
from dissect.ntfs.util import segment_reference

from dissect.target.exceptions import (
    FileNotFoundError,
    IsADirectoryError,
    NotADirectoryError,
    NotASymlinkError,
)
from dissect.target.filesystem import DirEntry, Filesystem, FilesystemEntry
from dissect.target.helpers import fsutil

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.ntfs.attr import FileName, StandardInformation
    from dissect.ntfs.util import AttributeMap


class NtfsFilesystem(Filesystem):
    __type__ = "ntfs"

    def __init__(
        self,
        fh: BinaryIO | None = None,
        boot: BinaryIO | None = None,
        mft: BinaryIO | None = None,
        usnjrnl: BinaryIO | None = None,
        sds: BinaryIO | None = None,
        *args,
        **kwargs,
    ):
        super().__init__(fh, *args, case_sensitive=False, alt_separator="\\", **kwargs)
        self.ntfs = NTFS(fh, boot=boot, mft=mft, usnjrnl=usnjrnl, sds=sds)

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        sector = fh.read(512)
        return sector[3:11] == NTFS_SIGNATURE

    def get(self, path: str) -> NtfsFilesystemEntry:
        return NtfsFilesystemEntry(self, path, self._get_record(path))

    def _get_record(self, path: str, root: MftRecord | None = None) -> MftRecord:
        try:
            path = path.rsplit(":", maxsplit=1)[0]
            return self.ntfs.mft.get(path, root=root)
        except NtfsFileNotFoundError as e:
            raise FileNotFoundError(path) from e
        except NtfsNotADirectoryError as e:
            raise NotADirectoryError(path) from e
        except NtfsError as e:
            raise FileNotFoundError(path) from e


class NtfsDirEntry(DirEntry):
    entry: IndexEntry

    def get(self) -> FilesystemEntry:
        return NtfsFilesystemEntry(self.fs, self.path, self.entry.dereference())

    def is_dir(self, *, follow_symlinks: bool = True) -> bool:
        if follow_symlinks and self.is_symlink():
            return super().is_dir(follow_symlinks=follow_symlinks)

        return self.entry.attribute.is_dir()

    def is_file(self, *, follow_symlinks: bool = True) -> bool:
        if follow_symlinks and self.is_symlink():
            return super().is_file(follow_symlinks=follow_symlinks)

        return self.entry.attribute.is_file()

    def is_symlink(self) -> bool:
        return self.entry.attribute.is_symlink() or self.entry.attribute.is_mount_point()

    def is_junction(self) -> bool:
        return self.entry.attribute.is_mount_point()

    def stat(self, *, follow_symlinks: bool = True) -> fsutil.stat_result:
        if follow_symlinks and self.is_symlink():
            return self.fs.stat(self.path)

        attr: FileName = self.entry.attribute
        if attr.is_symlink() or attr.is_mount_point():
            # Technically NTFS mount points/junctions are not symlinks, but it's easier if we pretend they are
            mode = stat.S_IFLNK
        elif attr.is_dir():
            mode = stat.S_IFDIR
        else:
            mode = stat.S_IFREG

        # mode, ino, dev, nlink, uid, gid, size, atime, mtime, ctime
        st_info = fsutil.stat_result(
            [
                mode | 0o777,
                segment_reference(self.entry.header.FileReference),
                id(self.fs),
                0,
                0,
                0,
                attr.file_size,
                attr.last_access_time.timestamp(),
                attr.last_modification_time.timestamp(),
                # ctime gets set to creation time for python <3.12 purposes
                attr.creation_time.timestamp(),
            ]
        )

        # Set the nanosecond resolution separately
        st_info.st_atime_ns = attr.last_access_time_ns
        st_info.st_mtime_ns = attr.last_modification_time_ns

        st_info.st_ctime_ns = attr.creation_time_ns

        st_info.st_birthtime = attr.creation_time.timestamp()
        st_info.st_birthtime_ns = attr.creation_time_ns

        # real_size is none if the size is resident
        st_info.st_blksize = self.entry.index.record.ntfs.cluster_size
        st_info.st_blocks = math.ceil(attr.allocated_size / 512)

        return st_info


class NtfsFilesystemEntry(FilesystemEntry):
    fs: NtfsFilesystem
    entry: MftRecord

    def __init__(self, fs: NtfsFilesystem, path: str, entry: MftRecord):
        super().__init__(fs, path, entry)

        self.ads = ""
        if ":" in self.path:
            self.path, self.ads = path.rsplit(":", maxsplit=1)

    def get(self, path: str) -> NtfsFilesystemEntry:
        return NtfsFilesystemEntry(
            self.fs,
            fsutil.join(self.path, path, alt_separator=self.fs.alt_separator),
            self.fs._get_record(path, self.entry),
        )

    def open(self, name: str = "") -> BinaryIO:
        if self.is_dir():
            raise IsADirectoryError(self.path)

        if self.is_symlink():
            return self.readlink_ext().open(name)

        stream = name or self.ads
        return self.entry.open(stream)

    def scandir(self, ignore_dos: bool = True) -> Iterator[NtfsDirEntry]:
        if not self.is_dir():
            raise NotADirectoryError(self.path)

        for entry in self._resolve().entry.iterdir(ignore_dos=ignore_dos):
            if entry.attribute.file_name == ".":
                continue

            yield NtfsDirEntry(self.fs, self.path, entry.attribute.file_name, entry)

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        if not follow_symlinks and self.is_symlink():
            return False

        return self.entry.is_dir()

    def is_file(self, follow_symlinks: bool = True) -> bool:
        if not follow_symlinks and self.is_symlink():
            return False

        return not self.is_dir(follow_symlinks=follow_symlinks)

    def is_symlink(self) -> bool:
        return self.entry.is_symlink() or self.entry.is_mount_point()

    def readlink(self) -> str:
        # Note: we only need to check and resolve symlinks when actually interacting with the target, such as
        # opening a file or iterating a directory. This is because the reparse point itself will have the appropriate
        # flags set to indicate if the target is a file or directory
        if not self.is_symlink():
            raise NotASymlinkError

        reparse_point = self.entry.attributes.REPARSE_POINT
        print_name = reparse_point.print_name
        if reparse_point.absolute:
            # Prefix with \\ to make the path play ball with all the filesystem utilities
            # Note: absolute links (such as directory junctions) will _always_ fail within the filesystem
            # This is because absolute links include the drive letter, of which we have no knowledge here
            # These will only work in the RootFilesystem
            print_name = "\\" + print_name
        return fsutil.normalize(print_name, self.fs.alt_separator)

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self._resolve(follow_symlinks=follow_symlinks).lstat()

    def lstat(self) -> fsutil.stat_result:
        record = self.entry

        size = 0
        real_size = 0
        if self.is_symlink():
            mode = stat.S_IFLNK
        elif self.is_file():
            mode = stat.S_IFREG
            try:
                size = record.size(self.ads)
                real_size = record.size(self.ads, allocated=True)
            except NtfsFileNotFoundError as e:
                # Occurs when it cannot find the the specific ads inside its attributes
                raise FileNotFoundError from e
        else:
            mode = stat.S_IFDIR

        stdinfo: StandardInformation = record.attributes.STANDARD_INFORMATION

        # mode, ino, dev, nlink, uid, gid, size, atime, mtime, ctime
        st_info = fsutil.stat_result(
            [
                mode | 0o777,
                record.segment,
                id(self.fs),
                record.header.ReferenceCount,
                0,
                0,
                size,
                stdinfo.last_access_time.timestamp(),
                stdinfo.last_modification_time.timestamp(),
                # ctime gets set to creation time for python <3.12 purposes
                stdinfo.creation_time.timestamp(),
            ]
        )

        # Set the nanosecond resolution separately
        st_info.st_atime_ns = stdinfo.last_access_time_ns
        st_info.st_mtime_ns = stdinfo.last_modification_time_ns

        st_info.st_ctime_ns = stdinfo.creation_time_ns

        st_info.st_birthtime = stdinfo.creation_time.timestamp()
        st_info.st_birthtime_ns = stdinfo.creation_time_ns

        # real_size is none if the size is resident
        st_info.st_blksize = record.ntfs.cluster_size
        blocks = 0
        if not record.resident:
            blocks = math.ceil(real_size / 512)

        st_info.st_blocks = blocks

        return st_info

    def attr(self) -> AttributeMap:
        if self.is_symlink():
            return self.readlink_ext().lattr()
        return self.lattr()

    def lattr(self) -> AttributeMap:
        return self.entry.attributes



from __future__ import annotations

import re
from typing import TYPE_CHECKING

import pytest
from flow.record.fieldtypes import datetime as dt

from dissect.target.helpers import keychain
from dissect.target.helpers.fsutil import TargetPath
from dissect.target.plugins.apps.browser.firefox import (
    FirefoxPlugin,
    decrypt_master_key,
    decrypt_value,
)
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


@pytest.fixture
def target_firefox_win(target_win_users: Target, fs_win: VirtualFilesystem) -> Target:
    fs_win.map_dir(
        "Users\\John\\AppData\\Local\\Mozilla\\Firefox\\Profiles\\g1rbw8y7.default-release\\",
        absolute_path("_data/plugins/apps/browser/firefox/generic/"),
    )

    target_win_users.add_plugin(FirefoxPlugin)

    return target_win_users


@pytest.fixture
def target_firefox_unix(target_unix_users: Target, fs_unix: VirtualFilesystem) -> Target:
    fs_unix.map_dir(
        "/root/.mozilla/firefox/g1rbw8y7.default-release/", absolute_path("_data/plugins/apps/browser/firefox/generic/")
    )

    target_unix_users.add_plugin(FirefoxPlugin)

    return target_unix_users


@pytest.fixture
def target_firefox_oculus(target_android: Target, fs_android: VirtualFilesystem) -> Target:
    fs_android.map_dir(
        "/data/data/org.mozilla.vrbrowser",
        absolute_path("_data/plugins/apps/browser/firefox/android/org.mozilla.vrbrowser"),
    )

    target_android.add_plugin(FirefoxPlugin)

    return target_android


@pytest.mark.parametrize(
    ("target_platform", "expected_source"),
    [
        (
            "target_firefox_win",
            "C:\\Users\\John\\AppData\\local\\Mozilla\\Firefox\\Profiles\\g1rbw8y7.default-release\\places.sqlite",
        ),
        (
            "target_firefox_unix",
            "/root/.mozilla/firefox/g1rbw8y7.default-release/places.sqlite",
        ),
    ],
)
def test_firefox_history(target_platform: str, expected_source: str, request: pytest.FixtureRequest) -> None:
    target: Target = request.getfixturevalue(target_platform)
    records = list(target.firefox.history())

    assert len(records) == 24
    assert {"firefox"} == {record.browser for record in records}

    assert records[0].ts == dt("2021-12-01T10:42:05.742000+00:00")
    assert records[0].browser == "firefox"
    assert records[0].id == 1
    assert records[0].url == "https://www.mozilla.org/privacy/firefox/"
    assert not records[0].title
    assert not records[0].description
    assert records[0].host == "www.mozilla.org"
    assert records[0].visit_type == 1
    assert records[0].visit_count == 1
    assert records[0].hidden
    assert not records[0].typed
    assert records[0].session == 0
    assert not records[0].from_visit
    assert not records[0].from_url
    assert records[0].source == expected_source


@pytest.mark.parametrize(
    "target_platform",
    ["target_firefox_win", "target_firefox_unix"],
)
def test_firefox_downloads(target_platform: str, request: pytest.FixtureRequest) -> None:
    target: Target = request.getfixturevalue(target_platform)
    records = list(target.firefox.downloads())

    assert len(records) == 3
    assert {"firefox"} == {record.browser for record in records}

    assert records[0].id == 1
    assert records[0].ts_start == dt("2021-12-01T10:57:01.175000+00:00")
    assert records[0].ts_end == dt("2021-12-01T10:57:01.321000+00:00")
    assert (
        records[0].url
        == "https://dl.google.com/tag/s/appguid%3D%7B8A69D345-D564-463C-AFF1-A69D9E530F96%7D%26iid%3D%7B2098EF96-29DB"
        "-B268-0B90-01AD59CD5C17%7D%26lang%3Dnl%26browser%3D3%26usagestats%3D1%26appname%3DGoogle%2520Chrome%26needs"
        "admin%3Dprefers%26ap%3Dx64-stable-statsdef_1%26installdataindex%3Dempty/update2/installers/ChromeSetup.exe"
    )


@pytest.mark.parametrize(
    "target_platform",
    ["target_firefox_win", "target_firefox_unix"],
)
def test_firefox_cookies(target_platform: str, request: pytest.FixtureRequest) -> None:
    target: Target = request.getfixturevalue(target_platform)

    records = list(target.firefox.cookies())

    assert len(records) == 4
    assert {"firefox"} == {record.browser for record in records}

    assert records[0].ts_created == dt("2023-07-13 09:53:47.460676+00:00")
    assert sorted(c.name for c in records) == [
        "_lr_env_src_ats",
        "_lr_retry_request",
        "_uc_referrer",
        "_uc_referrer",
    ]


@pytest.mark.parametrize(
    "target_platform",
    ["target_firefox_win", "target_firefox_unix"],
)
def test_firefox_extensions(target_platform: str, request: pytest.FixtureRequest) -> None:
    target: Target = request.getfixturevalue(target_platform)

    records = list(target.firefox.extensions())

    assert {"firefox"} == {record.browser for record in records}
    assert len(records) == 2
    assert records[0].extension_id == "uBlock0@raymondhill.net"
    assert records[0].ts_install == dt("2024-04-23 07:07:21.516000+00:00")
    assert records[0].ts_update == dt("2024-04-23 07:07:21.516000+00:00")
    assert (
        records[0].ext_path == "C:\\Users\\Win11\\AppData\\Roaming\\Mozilla\\Firefox\\Profiles"
        "\\9nxit8q0.default-release\\extensions\\uBlock0@raymondhill.net.xpi"
    )
    assert records[0].permissions == [
        "alarms",
        "dns",
        "menus",
        "privacy",
        "storage",
        "tabs",
        "unlimitedStorage",
        "webNavigation",
        "webRequest",
        "webRequestBlocking",
    ]
    assert records[1].permissions == []


@pytest.mark.parametrize(
    "target_platform",
    [
        "target_firefox_win",
        "target_firefox_unix",
    ],
)
def test_firefox_passwords(target_platform: str, request: pytest.FixtureRequest) -> None:
    target: Target = request.getfixturevalue(target_platform)

    records = list(target.firefox.passwords())
    assert len(records) == 2

    assert records[0].browser == "firefox"
    assert records[0].decrypted_username == "username"
    assert records[0].encrypted_password is None
    assert records[0].decrypted_password == "password"

    assert records[1].browser == "firefox"
    assert records[1].decrypted_username == "username"
    assert records[1].encrypted_password is None
    assert records[1].decrypted_password == "password"


def test_firefox_passwords_unix_primary_password(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_dir(
        "/root/.mozilla/firefox/g1rbw8y7.default-release/",
        absolute_path("_data/plugins/apps/browser/firefox/passwords/primary/"),
    )
    target_unix_users.add_plugin(FirefoxPlugin)

    keychain.register_key(
        keychain.KeyType.PASSPHRASE,
        "PrimaryPassword",
        identifier=None,
        provider="browser",
    )

    records = list(target_unix_users.firefox.passwords())

    assert len(records) == 1

    assert records[0].browser == "firefox"
    assert records[0].username == "root"
    assert records[0].user_home == "/root"
    assert records[0].decrypted_username == "username"
    assert records[0].encrypted_password is None
    assert records[0].decrypted_password == "password"


def test_firefox_history_oculus(target_firefox_oculus: Target) -> None:
    records = list(target_firefox_oculus.firefox.history())

    assert len(records) == 191
    assert {"firefox"} == {record.browser for record in records}

    assert records[0].url == "https://webxr.today/"
    assert records[0].id == 1
    assert not records[0].description
    assert records[0].ts == dt("2021-11-04 13:29:30.780000+00:00")


def test_firefox_cookies_oculus(target_firefox_oculus: Target) -> None:
    records = list(target_firefox_oculus.firefox.cookies())

    assert len(records) == 33
    assert {"firefox"} == {record.browser for record in records}

    assert records[0].ts_created == dt("2023-07-06 07:50:55.352988+00:00")
    assert sorted(c.name for c in records) == [
        "AEC",
        "AWSALBTGCORS",
        "CONSENT",
        "Features",
        "Locale",
        "NID",
        "OptanonConsent",
        "SNID",
        "SOCS",
        "Session-Id",
        "Tag",
        "Tag",
        "Tag",
        "Tag",
        "VISITOR_INFO1_LIVE",
        "_ga",
        "_ga_18EC69JQ0P",
        "_ga_2VC139B3XV",
        "_gat_gtag_UA_36116321_2",
        "_gid",
        "_hjFirstSeen",
        "count",
        "is-proton-user",
        "lst",
        "lst",
        "mp_abe3945ad0ddaadc3d987393d8d7c2ce_mixpanel",
        "session_id",
        "spid.6b32",
        "spses.6b32",
        "visid_incap_1329355",
        "visid_incap_2188750",
        "visid_incap_2295456",
        "visid_incap_2809573",
    ]


def test_firefox_extensions_oculus(target_firefox_oculus: Target) -> None:
    records = list(target_firefox_oculus.firefox.extensions())

    assert {"firefox"} == {record.browser for record in records}
    assert len(records) == 5
    assert records[0].extension_id == "default-theme@mozilla.org"
    assert records[0].ts_install == dt("2021-11-04 13:29:29.988000+00:00")
    assert records[0].ts_update == dt("1970-01-01 00:00:00+00:00")
    assert records[0].ext_path is None
    assert records[0].permissions == []


@pytest.fixture
def path_key4(fs_unix: VirtualFilesystem) -> TargetPath:
    fs_unix.map_file(
        "/key4.db",
        absolute_path("_data/plugins/apps/browser/firefox/generic/key4.db"),
    )
    return TargetPath(fs_unix, "/key4.db")


@pytest.fixture
def path_key4_primary_password(fs_unix: VirtualFilesystem) -> TargetPath:
    fs_unix.map_file(
        "/key4.db",
        absolute_path("_data/plugins/apps/browser/firefox/passwords/primary/key4.db"),
    )
    return TargetPath(fs_unix, "/key4.db")


def test_passwords(path_key4: TargetPath) -> None:
    """Test if we can decrypt a password entry with an empty Firefox primary password."""

    key = decrypt_master_key(path_key4, b"")

    b64_username = "MDoEEPgAAAAAAAAAAAAAAAAAAAEwFAYIKoZIhvcNAwcECIFVMX6MyYxpBBC2j8K+bCEaE9/FmqE1wo2A"
    b64_password = "MDoEEPgAAAAAAAAAAAAAAAAAAAEwFAYIKoZIhvcNAwcECB2ZPCZBORUJBBDZ8dBZUVgECoiFD5vPvTbP"

    plaintext_username = decrypt_value(b64_username, key)
    assert plaintext_username == b"username"

    plaintext_password = decrypt_value(b64_password, key)
    assert plaintext_password == b"password"


def test_passwords_primary_password(path_key4_primary_password: TargetPath) -> None:
    """Test if we can decrypt a password entry with a Firefox primary password."""

    key = decrypt_master_key(path_key4_primary_password, b"PrimaryPassword")

    b64_username = "MDoEEPgAAAAAAAAAAAAAAAAAAAEwFAYIKoZIhvcNAwcECPEZvPh6dhBkBBB3/O2puJy1NiBUo5gS8hZh"
    plaintext_username = decrypt_value(b64_username, key)
    assert plaintext_username == b"username"

    b64_password = "MDoEEPgAAAAAAAAAAAAAAAAAAAEwFAYIKoZIhvcNAwcECM00msnMxFyVBBByTarrnS+FSR5OHQhZfs8t"
    plaintext_password = decrypt_value(b64_password, key)
    assert plaintext_password == b"password"


def test_passwords_no_master_key() -> None:
    """Test if a fitting exception is raised when omitting a decryption key."""

    with pytest.raises(ValueError, match="Not a valid TDES key"):
        decrypt_value("MDoEEPgAAAAAAAAAAAAAAAAAAAEwFAYIKoZIhvcNAwcECIFVMX6MyYxpBBC2j8K+bCEaE9/FmqE1wo2A", b"")


def test_passwords_decrypt_master_key_invalid_primary_password(path_key4_primary_password: TargetPath) -> None:
    """Test if a fitting exception is raised when providing an invalid primary password."""

    with pytest.raises(ValueError, match=re.escape("Master key decryption failed.")):
        decrypt_master_key(path_key4_primary_password, b"BAD_PRIMARY_PASSWORD")


def test_passwords_decrypt_master_key(path_key4: TargetPath) -> None:
    """Test if we can decrypt a master key with no primary password."""

    key = decrypt_master_key(path_key4, b"")
    assert key.hex() == "452c2f920285f794614af1c2d99ed331940d73298a526140"


def test_passwords_decrypt_aes() -> None:
    """Test if we correctly decrypt Firefox 144.0 passwords. Data generated on a Windows 11 24H2 VM.

    ``key4.db`` file contains an empty entry in the ``nssPrivate`` table to test if we correctly grab
    the last row of the table when decrypting the master key.
    """

    key4_file = absolute_path("_data/plugins/apps/browser/firefox/passwords/144.0/key4.db")
    key = decrypt_master_key(key4_file, b"")

    b64_username = "MEMEEPgAAAAAAAAAAAAAAAAAAAEwHQYJYIZIAWUDBAEqBBC43iUHA33gEegQQOh2RYVABBDpNU6x1Q9IS20qq55IPy8K"
    plaintext_username = decrypt_value(b64_username, key)
    assert plaintext_username == b"username"

    b64_password = "MEMEEPgAAAAAAAAAAAAAAAAAAAEwHQYJYIZIAWUDBAEqBBBnaT+EZVKmFCEX256l7Z06BBDuNPzjclPu77lO5m8ZC2HP"
    plaintext_password = decrypt_value(b64_password, key)
    assert plaintext_password == b"password"

    b64_unknown = "MEMEEPgAAAAAAAAAAAAAAAAAAAEwHQYJYIZIAWUDBAEqBBDH5u9D6iZZ7WbagSHJpDGnBBCeA5eSi8teghT3FHogzqOF"
    assert decrypt_value(b64_unknown, key) == b""


def test_passwords_backup(target_win_users: Target, fs_win: VirtualFilesystem) -> None:
    """Test if we find ``logins-backup.json`` files."""

    fs_win.map_dir(
        "Users/John/AppData/Roaming/Mozilla/Firefox/Profiles/8jb1c7qs.default-release",
        absolute_path("_data/plugins/apps/browser/firefox/passwords/144.0/"),
    )
    target_win_users.add_plugin(FirefoxPlugin)

    records = list(target_win_users.firefox.passwords())
    assert len(records) == 3

    assert (
        records[-1].source
        == "C:\\Users\\John\\AppData\\Roaming\\Mozilla\\Firefox\\Profiles\\8jb1c7qs.default-release\\logins-backup.json"
    )


def test_firefox_linux_xdg_config_home(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if we find Mozilla Firefox Linux default XDG_CONFIG_HOME locations."""

    fs_unix.map_dir(
        "/home/user/.config/mozilla/firefox/g1rbw8y7.Profile Name",
        absolute_path("_data/plugins/apps/browser/firefox/generic"),
    )

    plugin = target_unix_users.add_plugin(FirefoxPlugin)
    assert str(plugin.installs[0][1]) == "/home/user/.config/mozilla/firefox"



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.helpers.certificate import COMMON_CERTIFICATE_FIELDS
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import NamespacePlugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

WebserverAccessLogRecord = TargetRecordDescriptor(
    "application/webserver/log/access",
    [
        ("datetime", "ts"),
        ("string", "webserver"),
        ("string", "remote_user"),
        ("net.ipaddress", "remote_ip"),
        ("net.ipaddress", "local_ip"),
        ("varint", "pid"),
        ("string", "method"),
        ("uri", "uri"),
        ("string", "query"),
        ("string", "protocol"),
        ("varint", "status_code"),
        ("varint", "bytes_sent"),
        ("uri", "referer"),
        ("string", "useragent"),
        ("varint", "response_time_ms"),
        ("path", "source"),
    ],
)

WebserverErrorLogRecord = TargetRecordDescriptor(
    "application/webserver/log/error",
    [
        ("datetime", "ts"),
        ("string", "webserver"),
        ("net.ipaddress", "remote_ip"),
        ("varint", "pid"),
        ("string", "module"),
        ("string", "level"),
        ("string", "error_source"),
        ("string", "error_code"),
        ("string", "message"),
        ("path", "source"),
    ],
)

WebserverHostRecord = TargetRecordDescriptor(
    "application/webserver/host",
    [
        ("datetime", "ts"),
        ("string", "webserver"),
        ("string", "server_name"),
        ("varint", "server_port"),
        ("path", "root_path"),
        ("path", "access_log_config"),
        ("path", "error_log_config"),
        ("path", "tls_certificate"),
        ("path", "tls_key"),
        ("path", "source"),
    ],
)

WebserverCertificateRecord = TargetRecordDescriptor(
    "application/webserver/host/certificate",
    [
        ("datetime", "ts"),
        ("string", "webserver"),
        *COMMON_CERTIFICATE_FIELDS,
        ("string", "host"),
        ("path", "source"),
    ],
)


class WebserverPlugin(NamespacePlugin):
    __namespace__ = "webserver"

    @export(record=[WebserverAccessLogRecord, WebserverErrorLogRecord])
    def logs(self) -> Iterator[WebserverAccessLogRecord | WebserverErrorLogRecord]:
        """Returns log file records from installed webservers."""
        yield from self.access()
        yield from self.error()



from __future__ import annotations

import re
from enum import Enum, auto
from typing import TYPE_CHECKING

from dissect.ntfs.exceptions import FileNotFoundError

if TYPE_CHECKING:
    from dissect.ntfs.mft import MftRecord

    from dissect.target.filesystems.ntfs import NtfsFilesystem
    from dissect.target.target import Target

DRIVE_LETTER_RE = re.compile(r"[a-zA-Z]:")


class InformationType(Enum):
    """Valid information types."""

    STANDARD_INFORMATION = auto()
    FILE_INFORMATION = auto()
    ALTERNATE_DATA_STREAM = auto()


def get_drive_letter(target: Target, filesystem: NtfsFilesystem) -> str:
    """Retrieve the drive letter from the loaded mounts

    When the drive letter is not available for that filesystem it returns empty.
    """
    # A filesystem can be known under multiple drives (mount points). If it is
    # a windows system volume, there are the default sysvol and c: drives.
    # If the target has a virtual ntfs filesystem, e.g. as constructed by the
    # tar and dir loaders, there is also the /$fs$/fs<n> drive, under which the
    # "fake" ntfs filesystem is mounted.
    # The precedence for drives is first the drive letter drives (e.g. c:),
    # second the "normally" named drives (e.g. sysvol) and finally the anonymous
    # drives (e.g. /$fs/fs0).
    mount_items = (item for item in target.fs.mounts.items() if hasattr(item[1], "ntfs"))
    drives = [key for key, fs in mount_items if fs.ntfs is filesystem.ntfs]

    single_letter_drives = []
    other_drives = []
    anon_drives = []

    for drive in drives:
        if DRIVE_LETTER_RE.match(drive):
            single_letter_drives.append(drive)
        elif "$fs$" in drive:
            anon_drives.append(drive)
        else:
            other_drives.append(drive)

    drives = sorted(single_letter_drives) + sorted(other_drives) + sorted(anon_drives)

    if drives:
        return f"{drives[0]}\\"
    return ""


def get_volume_identifier(fs: NtfsFilesystem) -> str | None:
    """Return the filesystem GUID or serial, if available."""

    if fs.volume and (guid := getattr(fs.volume, "guid", None)):
        return guid

    # Fallback to NTFS serial if volume GUID is not available
    if fs.ntfs.serial:
        return str(fs.ntfs.serial)

    return None


def get_owner_and_group(entry: MftRecord, fs: NtfsFilesystem) -> tuple[str | None, str | None]:
    owner, group = None, None
    try:
        stdinfo = entry.attributes.STANDARD_INFORMATION[0]
        sd = fs.ntfs.secure.lookup(stdinfo.security_id)
        owner, group = sd.owner, sd.group
    except (AttributeError, IndexError, KeyError):
        # When $STANDARD_INFORMATION cannot be found or no $Secure file is loaded
        pass
    return owner, group


def get_record_size(record: MftRecord, name: str = "") -> int | None:
    """Gets the size for a specific record."""
    try:
        return record.size(name)
    except (FileNotFoundError, KeyError):
        # FileNotFoundError is raised when the entry has no attributes
        # KeyError is raised when it tries to access a specific attribute that does not exist
        pass
    return None



from __future__ import annotations

import base64
import binascii
from hashlib import md5, sha1, sha256

from dissect.cstruct import cstruct

from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import NamespacePlugin

rfc4716_def = """
struct ssh_string {
    uint32 length;
    char value[length];
}

struct ssh_private_key {
    char magic[15];

    ssh_string cipher;
    ssh_string kdf_name;
    ssh_string kdf_options;

    uint32 number_of_keys;

    ssh_string public;
    ssh_string private;
}
"""

c_rfc4716 = cstruct(endian=">").load(rfc4716_def)

RFC4716_MARKER_START = b"-----BEGIN OPENSSH PRIVATE KEY-----"
RFC4716_MARKER_END = b"-----END OPENSSH PRIVATE KEY-----"
RFC4716_MAGIC = b"openssh-key-v1\x00"
RFC4716_PADDING = b"\x01\x02\x03\x04\x05\x06\x07"
RFC4716_NONE = b"none"

PKCS8_MARKER_START = b"-----BEGIN PRIVATE KEY-----"
PKCS8_MARKER_END = b"-----END PRIVATE KEY-----"
PKCS8_MARKER_START_ENCRYPTED = b"-----BEGIN ENCRYPTED PRIVATE KEY-----"
PKCS8_MARKER_END_ENCRYPTED = b"-----END ENCRYPTED PRIVATE KEY-----"

PEM_MARKER_START_RSA = b"-----BEGIN RSA PRIVATE KEY-----"
PEM_MARKER_END_RSA = b"-----END RSA PRIVATE KEY-----"
PEM_MARKER_START_DSA = b"-----BEGIN DSA PRIVATE KEY-----"
PEM_MARKER_END_DSA = b"-----END DSA PRIVATE KEY-----"
PEM_MARKER_START_EC = b"-----BEGIN EC PRIVATE KEY-----"
PEM_MARKER_END_EC = b"-----END EC PRIVATE KEY-----"
PEM_ENCRYPTED = b"ENCRYPTED"


OpenSSHUserRecordDescriptor = create_extended_descriptor([UserRecordDescriptorExtension])

COMMON_ELLEMENTS = [
    ("string", "key_type"),
    ("string", "comment"),
    ("path", "path"),
]

AuthorizedKeysRecord = OpenSSHUserRecordDescriptor(
    "application/openssh/authorized_keys",
    [
        *COMMON_ELLEMENTS,
        ("bytes", "public_key_pem"),
        ("digest", "fingerprint"),
        ("string", "options"),
    ],
)


KnownHostRecord = OpenSSHUserRecordDescriptor(
    "application/openssh/known_host",
    [
        ("datetime", "mtime_ts"),
        *COMMON_ELLEMENTS,
        ("string", "host"),
        ("varint", "port"),
        ("bytes", "public_key_pem"),
        ("string", "marker"),
        ("digest", "fingerprint"),
    ],
)


PrivateKeyRecord = OpenSSHUserRecordDescriptor(
    "application/openssh/private_key",
    [
        ("datetime", "mtime_ts"),
        *COMMON_ELLEMENTS,
        ("string", "key_format"),
        ("bytes", "private_key_pem"),
        ("bytes", "public_key_pem"),
        ("digest", "fingerprint"),
        ("boolean", "encrypted"),
    ],
)

PublicKeyRecord = OpenSSHUserRecordDescriptor(
    "application/openssh/public_key",
    [
        ("datetime", "mtime_ts"),
        *COMMON_ELLEMENTS,
        ("bytes", "public_key_pem"),
        ("digest", "fingerprint"),
    ],
)


class SSHPlugin(NamespacePlugin):
    __namespace__ = "ssh"


def calculate_fingerprints(public_key_decoded: bytes, ssh_keygen_format: bool = False) -> tuple[str, str, str]:
    """Calculate the MD5, SHA1 and SHA256 digest of the given decoded public key.

    Adheres as much as possible to the output provided by ssh-keygen when ``ssh_keygen_format``
    parameter is set to ``True``. When set to ``False`` (default) hexdigests are calculated
    instead for ``sha1``and ``sha256``.

    References:
        - https://en.wikipedia.org/wiki/Public_key_fingerprint
        - https://man7.org/linux/man-pages/man1/ssh-keygen.1.html
        - ``ssh-keygen -l -E <alg> -f key.pub``
    """
    if not public_key_decoded:
        raise ValueError("No decoded public key provided")

    if not isinstance(public_key_decoded, bytes):
        raise TypeError("Provided public key should be bytes")

    if public_key_decoded[0:3] != b"\x00\x00\x00":
        raise ValueError("Provided value does not look like a public key")

    digest_md5 = md5(public_key_decoded).digest()
    digest_sha1 = sha1(public_key_decoded).digest()
    digest_sha256 = sha256(public_key_decoded).digest()

    if ssh_keygen_format:
        fingerprint_sha1 = base64.b64encode(digest_sha1).rstrip(b"=").decode()
        fingerprint_sha256 = base64.b64encode(digest_sha256).rstrip(b"=").decode()
    else:
        fingerprint_sha1 = digest_sha1.hex()
        fingerprint_sha256 = digest_sha256.hex()

    return digest_md5.hex(), fingerprint_sha1, fingerprint_sha256


def is_rfc4716(data: bytes) -> bool:
    """Validate data is a valid looking SSH private key in the OpenSSH format."""
    return data.startswith(RFC4716_MARKER_START) and data.endswith(RFC4716_MARKER_END)


def decode_rfc4716(data: bytes) -> bytes:
    """Base64 decode the private key data."""
    encoded_key_data = data.removeprefix(RFC4716_MARKER_START).removesuffix(RFC4716_MARKER_END)
    try:
        return base64.b64decode(encoded_key_data)
    except binascii.Error:
        raise ValueError("Error decoding RFC4716 key data")


def is_pkcs8(data: bytes) -> bool:
    """Validate data is a valid looking PKCS8 SSH private key."""
    return (data.startswith(PKCS8_MARKER_START) and data.endswith(PKCS8_MARKER_END)) or (
        data.startswith(PKCS8_MARKER_START_ENCRYPTED) and data.endswith(PKCS8_MARKER_END_ENCRYPTED)
    )


def is_pem(data: bytes) -> bool:
    """Validate data is a valid looking PEM SSH private key."""
    return (
        (data.startswith(PEM_MARKER_START_RSA) and data.endswith(PEM_MARKER_END_RSA))
        or (data.startswith(PEM_MARKER_START_DSA) and data.endswith(PEM_MARKER_END_DSA))
        or (data.startswith(PEM_MARKER_START_EC) and data.endswith(PEM_MARKER_END_EC))
    )


class SSHPrivateKey:
    """A class to parse (OpenSSH-supported) SSH private keys.

    OpenSSH supports three types of keys:
    * RFC4716 (default)
    * PKCS8
    * PEM
    """

    def __init__(self, data: bytes):
        self.key_type = None
        self.public_key = None
        self.comment = ""

        if is_rfc4716(data):
            self.format = "RFC4716"
            self._parse_rfc4716(data)

        elif is_pkcs8(data):
            self.format = "PKCS8"
            self.is_encrypted = data.startswith(PKCS8_MARKER_START_ENCRYPTED)

        elif is_pem(data):
            self.format = "PEM"
            self._parse_pem(data)

        else:
            raise ValueError("Unsupported private key format")

    def _parse_rfc4716(self, data: bytes) -> None:
        """Parse OpenSSH format SSH private keys.

        The format:
        "openssh-key-v1"0x00    # NULL-terminated "Auth Magic" string
        32-bit length, "none"   # ciphername length and string
        32-bit length, "none"   # kdfname length and string
        32-bit length, nil      # kdf (0 length, no kdf)
        32-bit 0x01             # number of keys, hard-coded to 1 (no length)
        32-bit length, sshpub   # public key in ssh format
            32-bit length, keytype
            32-bit length, pub0
            32-bit length, pub1
        32-bit length for rnd+prv+comment+pad
            64-bit dummy checksum?  # a random 32-bit int, repeated
            32-bit length, keytype  # the private key (including public)
            32-bit length, pub0     # Public Key parts
            32-bit length, pub1
            32-bit length, prv0     # Private Key parts
            ...                     # (number varies by type)
            32-bit length, comment  # comment string
            padding bytes 0x010203  # pad to blocksize (see notes below)

        Source: https://coolaj86.com/articles/the-openssh-private-key-format/
        """

        key_data = decode_rfc4716(data)
        private_key = c_rfc4716.ssh_private_key(key_data)

        # RFC4716 only supports 1 key at the moment.
        if private_key.magic != RFC4716_MAGIC or private_key.number_of_keys != 1:
            raise ValueError("Unexpected number of keys for RFC4716 format private key")

        self.is_encrypted = private_key.cipher.value != RFC4716_NONE

        self.public_key = base64.b64encode(private_key.public.value)
        public_key_type = c_rfc4716.ssh_string(private_key.public.value)
        self.key_type = public_key_type.value

        if not self.is_encrypted:
            private_key_data = private_key.private.value.rstrip(RFC4716_PADDING)

            # We skip the two dummy uint32s at the start.
            private_key_index = 8

            private_key_type = c_rfc4716.ssh_string(private_key_data[private_key_index:])
            private_key_index += 4 + private_key_type.length
            self.key_type = private_key_type.value

            private_key_fields = []
            while private_key_index < len(private_key_data):
                field = c_rfc4716.ssh_string(private_key_data[private_key_index:])
                private_key_index += 4 + field.length
                private_key_fields.append(field)

            # There is always a comment present (with a length field of 0 for empty comments).
            self.comment = private_key_fields[-1].value

    def _parse_pem(self, data: bytes) -> None:
        """Detect key type and encryption of PEM keys."""
        self.is_encrypted = PEM_ENCRYPTED in data

        if data.startswith(PEM_MARKER_START_RSA):
            self.key_type = "ssh-rsa"

        elif data.startswith(PEM_MARKER_START_DSA):
            self.key_type = "ssh-dss"

        # This is not a valid SSH key type, but we currently do not detect the specific ecdsa variant.
        else:
            self.key_type = "ecdsa"



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.plugins.child.qemu import QemuChildTargetPlugin
from tests.conftest import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_compatible(target_linux: Target, fs_linux: VirtualFilesystem) -> None:
    with pytest.raises(UnsupportedPluginError):
        QemuChildTargetPlugin(target_linux).check_compatible()

    fs_linux.map_file(
        "/etc/libvirt/qemu/qemu.xml",
        absolute_path("_data/loaders/libvirt/qemu.xml"),
    )
    QemuChildTargetPlugin(target_linux).check_compatible()


def test_list_children(target_linux: Target, fs_linux: VirtualFilesystem) -> None:
    qemu_xml = absolute_path("_data/loaders/libvirt/qemu.xml")
    fs_linux.map_file("/etc/libvirt/qemu/linux2022.xml", qemu_xml)
    fs_linux.map_file("/etc/libvirt/qemu/linux2024.xml", qemu_xml)

    child_plugin = QemuChildTargetPlugin(target_linux)

    children = list(child_plugin.list_children())

    assert len(children) == 2

    child = children[0]
    assert child.type == "qemu"
    assert child.name == "linux2022"
    assert child.path == "/etc/libvirt/qemu/linux2022.xml"



from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING, BinaryIO

import pytest

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.helpers import magic
from tests._utils import absolute_path

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.filesystem import FilesystemEntry


fs = VirtualFilesystem()
fs.map_file_fh("example.png", BytesIO(b""))
entry = fs.get("example.png")


@pytest.mark.parametrize(
    ("func", "input", "mime_out", "expected_output"),
    [
        # Archives
        ("from_buffer", bytes.fromhex("1f9d"), True, "application/x-compress"),
        ("from_buffer", bytes.fromhex("1fa0"), True, "application/x-compress"),
        ("from_buffer", bytes.fromhex("00002D6C68302D"), True, "application/x-lha"),
        ("from_buffer", bytes.fromhex("00002D6C68352D"), True, "application/x-lha"),
        ("from_buffer", bytes.fromhex("edabeedb"), True, "application/x-rpm"),
        ("from_buffer", b"BZh", True, "application/x-bzip2"),
        ("from_buffer", b"LZIP", True, "application/x-lzip"),
        ("from_buffer", b"070701", True, "application/x-cpio"),
        ("from_buffer", b"070702", True, "application/x-cpio"),
        ("from_buffer", b"070707", True, "application/x-cpio"),
        ("from_buffer", bytes.fromhex("504b0304"), True, "application/zip"),
        ("from_buffer", bytes.fromhex("504b0506"), True, "application/zip"),
        ("from_buffer", bytes.fromhex("504b0708"), True, "application/zip"),
        ("from_buffer", bytes.fromhex("526172211a07"), True, "application/vnd.rar"),
        ("from_buffer", bytes.fromhex("526172211a070100"), True, "application/vnd.rar"),
        ("from_buffer", bytes.fromhex("0E031301"), True, "application/x-hdf"),
        ("from_buffer", bytes.fromhex("894844460D0A1A0A"), True, "application/x-hdf"),
        pytest.param("from_buffer", (b"\x00" * 0x8001) + b"CD001", True, "application/vnd.efi.iso", id="iso"),
        ("from_buffer", b"xar!", True, "application/x-xar"),
        pytest.param("from_buffer", (b"\x00" * 257) + b"ustar\x00\x30\x30", False, "Tar archive", id="tar1"),
        pytest.param("from_buffer", (b"\x00" * 257) + b"ustar\x20\x20\x00", False, "Tar archive", id="tar2"),
        ("from_buffer", b"\x37\x7a\xbc\xaf\x27\x1c", True, "application/x-7z-compressed"),
        ("from_buffer", b"\x1f\x8b", True, "application/gzip"),
        ("from_buffer", b"\xfd\x37\x7a\x58\x5a\x00", True, "application/x-xz"),
        ("from_buffer", b"\x04\x22\x4d\x18", True, "application/x-lz4"),
        ("from_buffer", b"MSCF\x00\x00\x00\x00", True, "application/vnd.ms-cab-compressed"),
        ("from_buffer", b"\x78\x01", True, "application/zlib"),
        ("from_buffer", b"\x78\x5e", True, "application/zlib"),
        ("from_buffer", b"\x78\x9c", True, "application/zlib"),
        ("from_buffer", b"\x78\xda", True, "application/zlib"),
        ("from_buffer", b"\x78\x20", True, "application/zlib"),
        ("from_buffer", b"\x78\x7d", True, "application/zlib"),
        ("from_buffer", b"\x78\xbb", True, "application/zlib"),
        ("from_buffer", b"\x78\xf9", True, "application/zlib"),
        ("from_buffer", bytes.fromhex("4f626a01"), True, "application/avro"),
        ("from_buffer", bytes.fromhex("28B52FFD"), True, "application/zstd"),
        ("from_buffer", b"IsZ!", True, "application/vnd.efi.iso+compressed"),
        ("from_buffer", b"TAPE", True, "application/vnd.ms-tape"),
        # Database formats
        ("from_buffer", b"SQLite format 3\x00FILE DATA\x4d\x3c\xb2\xa1", True, "application/vnd.sqlite3"),
        ("from_buffer", b"SQLite format 3\x00FILE DATA\x4d\x3c\xb2\xa1", None, "SQLite3 database"),
        ("from_buffer", b"DUCK", True, "application/x-duckdb"),
        # Images
        ("from_buffer", b"\x00\x00\x01\x00", True, "image/vnd.microsoft.icon"),
        ("from_buffer", b"icns", True, "image/x-icns"),
        ("from_buffer", (b"\x00" * 4) + b"ftypheic", True, "image/heif"),
        ("from_buffer", b"GIF87a", True, "image/gif"),
        ("from_buffer", b"GIF89a", True, "image/gif"),
        ("from_buffer", bytes.fromhex("49492a00"), True, "image/tiff"),
        ("from_buffer", bytes.fromhex("4d4d002a"), True, "image/tiff"),
        ("from_buffer", bytes.fromhex("49492B00"), True, "image/bigtiff"),
        ("from_buffer", bytes.fromhex("4D4D002B"), True, "image/bigtiff"),
        ("from_buffer", bytes.fromhex("49492A00100000004352"), True, "image/tiff"),
        ("from_buffer", bytes.fromhex("425047FB"), True, "image/bpg"),
        ("from_buffer", bytes.fromhex("ffd8ffdb010203"), True, "image/jpeg"),
        ("from_buffer", bytes.fromhex("ffd8ffdb010203"), False, "JPEG image"),
        ("from_buffer", bytes.fromhex("FFD8FFE000104A4649460001"), True, "image/jpeg"),
        ("from_buffer", bytes.fromhex("FFD8FFEE"), True, "image/jpeg"),
        ("from_buffer", bytes.fromhex("FFD8FFE1"), True, "image/jpeg"),
        ("from_buffer", bytes.fromhex("FFD8FFE0"), True, "image/jpeg"),
        ("from_buffer", bytes.fromhex("ffd8ffdb"), True, "image/jpeg"),
        ("from_buffer", bytes.fromhex("ffd8ffe0"), True, "image/jpeg"),
        ("from_buffer", bytes.fromhex("ffd8ffe1"), True, "image/jpeg"),
        ("from_buffer", bytes.fromhex("0000000C6A5020200D0A870A"), True, "image/jp2"),
        ("from_buffer", bytes.fromhex("FF4FFF51"), True, "image/x-jp2-codestream"),
        ("from_buffer", bytes.fromhex("89504e470d0a1a0a"), True, "image/png"),
        ("from_buffer", b"8BPS  \x00\x00\x00\x00", True, "image/vnd.adobe.photoshop"),
        # Audio and video
        ("from_buffer", b"OggS", True, "application/ogg"),
        ("from_buffer", bytes.fromhex("52494646"), True, "application/x-riff"),
        ("from_buffer", bytes.fromhex("fffb"), True, "audio/mpeg"),
        ("from_buffer", bytes.fromhex("fff3"), True, "audio/mpeg"),
        ("from_buffer", bytes.fromhex("fff2"), True, "audio/mpeg"),
        ("from_buffer", b"ID3", True, "audio/mpeg"),
        ("from_buffer", b"fLaC", True, "audio/flac"),
        ("from_buffer", b"MThd", True, "audio/midi"),
        ("from_buffer", bytes.fromhex("1A45DFA3"), True, "application/x-matroska"),
        ("from_buffer", b"\x00\x00\x00\x00ftypisom", True, "video/mp4"),
        ("from_buffer", b"\x00\x00\x00\x00ftypMSNV", True, "video/mp4"),
        ("from_buffer", b"#EXTM3U", True, "audio/x-mpegurl"),
        # Productivity
        ("from_buffer", b"\x25\x50\x44\x46-", True, "application/pdf"),
        ("from_buffer", b"\xd0\xcf\x11\xe0\xa1\xb1\x1a\xe1", True, "application/vnd.ms-excel"),
        ("from_buffer", b"Received:", True, "message/rfc822"),
        ("from_buffer", b"!BDN", True, "application/vnd.ms-outlook-pst"),
        # Executables
        ("from_buffer", b"MZ", True, "application/x-dosexec"),
        ("from_buffer", bytes.fromhex("7F454C46"), True, "application/x-executable"),
        ("from_buffer", bytes.fromhex("cafebabe"), True, "application/x-java"),
        ("from_buffer", bytes.fromhex("feedface"), True, "application/x-mach-o"),
        ("from_buffer", bytes.fromhex("feedfacf"), True, "application/x-mach-o"),
        ("from_buffer", bytes.fromhex("feedfeed"), True, "application/x-java-keystore"),
        ("from_buffer", bytes.fromhex("cefaedfe"), True, "application/x-mach-o"),
        ("from_buffer", bytes.fromhex("cffaedfe"), True, "application/x-mach-o"),
        ("from_buffer", b"%!PS", True, "application/postscript"),
        ("from_buffer", b"\x00asm", True, "application/wasm"),
        ("from_buffer", b"!<arch>\x0a", True, "application/vnd.debian.binary-package"),
        ("from_buffer", bytes.fromhex("1B4C7561"), True, "application/x-lua"),
        # Network
        ("from_buffer", bytes.fromhex("d4c3b2a1"), True, "application/vnd.tcpdump.pcap"),
        ("from_buffer", bytes.fromhex("a1b2c3d4"), True, "application/vnd.tcpdump.pcap"),
        ("from_buffer", bytes.fromhex("4d3cb2a1"), True, "application/vnd.tcpdump.pcap"),
        ("from_buffer", bytes.fromhex("a1b23c4d"), True, "application/vnd.tcpdump.pcap"),
        ("from_buffer", bytes.fromhex("0a0d0d0a"), True, "application/x-pcapng"),
        # Certificates
        ("from_buffer", b"-----BEGIN CERTIFICATE-----", True, "application/pkix-cert"),
        ("from_buffer", b"-----BEGIN CERTIFICATE REQUEST-----", True, "text/pkix-csr"),
        ("from_buffer", b"-----BEGIN PRIVATE KEY-----", True, "text/x-ssh-private-key"),
        ("from_buffer", b"-----BEGIN DSA PRIVATE KEY-----", True, "text/x-ssh-private-key"),
        ("from_buffer", b"-----BEGIN RSA PRIVATE KEY-----", True, "text/x-ssh-private-key"),
        ("from_buffer", b"PuTTY-User-Key-File-2:", True, "text/x-putty-private-key"),
        ("from_buffer", b"PuTTY-User-Key-File-3:", True, "text/x-putty-private-key"),
        ("from_buffer", b"-----BEGIN OPENSSH PRIVATE KEY-----", True, "text/x-ssh-private-key"),
        ("from_buffer", b"-----BEGIN SSH2 PUBLIC KEY-----", True, "text/x-ssh-public-key"),
        ("from_buffer", b"-----BEGIN PGP PUBLIC KEY BLOCK-----", True, "application/pgp-keys"),
        # Web
        ("from_buffer", b"wOFF", True, "font/woff"),
        ("from_buffer", b"wOF2", True, "font/woff2"),
        ("from_buffer", b"ITSF\x03\x00\x00\x00\x60\x00\x00\x00", True, "application/vnd.ms-htmlhelp"),
        # Containers and disks
        ("from_buffer", b"KDMV\x01\x00\x00\x00", True, "application/x-vmdk-disk"),
        ("from_buffer", b"AFF", True, "application/x-aff"),
        ("from_buffer", b"EVF2", True, "application/x-encase"),
        ("from_buffer", b"EVF", True, "application/x-encase"),
        ("from_buffer", b"QFI", True, "application/x-qemu-disk"),
        ("from_buffer", b"<<< Oracle VM VirtualBox Disk Image >>>", True, "application/x-oracle-virtualbox-vdi"),
        ("from_buffer", b"conectix", True, "application/x-vhd-disk"),
        ("from_buffer", b"vhdxfile", True, "application/x-vhdx-disk"),
        ("from_buffer", b"MSWIM\x00\x00\x00\xd0\x00\x00\x00\x00", True, "application/x-ms-wim"),
        # Log files
        ("from_buffer", b"LfLe", True, "application/x-win-evt"),
        ("from_buffer", b"ElfFile", True, "application/x-win-evtx"),
        ("from_buffer", b"regf", True, "application/x-win-regf"),
        ("from_buffer", b"[ZoneTransfer]", True, "application/x-win-zonetransfer"),
        ("from_buffer", b"LPKSHHRH", True, "application/x-systemd-journald-log"),
        ("from_buffer", b"\xc0\x3b\x39\x98", True, "application/x-ext-jbd-fs-journal"),
        # Other data formats
        ("from_buffer", b"bplist", True, "application/x-plist"),
        ("from_buffer", b"<?xml ", True, "application/xml"),
        # Negatives
        ("from_buffer", b"unknown_file", None, None),
        ("from_buffer", b"", None, None),
        # Other methods
        ("from_file", absolute_path("_data/filesystems/cpio/initrd.img-6.1.0-15-amd64"), True, "application/zstd"),
        ("from_entry", entry, True, "image/png"),
        ("from_descriptor", BytesIO(b"ElfFile"), True, "application/x-win-evtx"),
        ("from_fh", BytesIO(b"regf"), True, "application/x-win-regf"),
    ],
)
def test_magic_detection(
    func: str, input: bytes | BinaryIO | Path | FilesystemEntry | None, mime_out: bool, expected_output: str
) -> None:
    """Test if we correctly identify common files.

    Relevant examples taken from referenced wikipedia page, grouped by types.

    References:
        - https://en.wikipedia.org/wiki/List_of_file_signatures
    """

    assert getattr(magic, func)(input, mime=mime_out) == expected_output


def test_magic_exception_handling() -> None:
    """Test if we throw sensible exception messages."""

    with pytest.raises(TypeError, match="Provided path is not a Path instance"):
        magic.from_file("not a Path")

    with pytest.raises(TypeError, match="Provided buf is not bytes"):
        magic.from_buffer("string")

    with pytest.raises(TypeError, match="Provided fh does not have a read or seek method"):
        magic.from_descriptor(b"not a buffer")

    with pytest.raises(TypeError, match="Provided suffix is not a string"):
        magic.Magic.detect(b"foo", suffix=b"not a string", mime=False)



from __future__ import annotations

import io
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import Mock, mock_open, patch

import pytest

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugins.os.windows.recyclebin import RecyclebinPlugin, c_recyclebin

if TYPE_CHECKING:
    from dissect.target.target import Target


@pytest.fixture
def recycle_bin(tmp_path: Path) -> VirtualFilesystem:
    recycle_bin = VirtualFilesystem()
    recycle_bin.map_dir("$a_random_recycle_bin_file", tmp_path)
    return recycle_bin


def test_recycle_bin_compatibility_failed(target_win: Target) -> None:
    with pytest.raises(UnsupportedPluginError):
        RecyclebinPlugin(target_win).check_compatible()


def test_recycle_bin_compat_succeeded(target_win: Target, recycle_bin: VirtualFilesystem) -> None:
    target_win.fs.mount("C:\\$recycle.bin", recycle_bin)
    assert RecyclebinPlugin(target_win).check_compatible() is None


def test_read_recycle_bin(target_win: Target) -> None:
    mocked_file = Mock()

    mocked_file.is_file.return_value = True
    mocked_file.is_dir.return_value = False

    test_file = Mock()
    with patch.object(RecyclebinPlugin, "read_bin_file", return_value=[test_file]) as mocked_bin_file:
        assert [test_file] == list(RecyclebinPlugin(target_win).read_recycle_bin(mocked_file))
        mocked_bin_file.assert_called_once_with(mocked_file)


def test_filtered_name(target_win: Target) -> None:
    mocked_file = Mock()
    mocked_file.is_file.return_value = True
    mocked_file.is_dir.return_value = False

    mocked_file.name = "hello"

    assert list(RecyclebinPlugin(target_win).read_recycle_bin(mocked_file)) == []


def test_read_recycle_bin_directory(target_win: Target) -> None:
    mocked_dir = Mock()
    mocked_dir.is_file.return_value = False
    mocked_dir.is_dir.return_value = True
    mocked_dir.name = "SID-SOME_RANDOM_SID"

    mocked_file = Mock()
    mocked_file.is_file.return_value = True
    mocked_file.is_dir.return_value = False
    mocked_file.name = "$ihello"

    mocked_dir.iterdir.return_value = [mocked_file] * 3

    with patch.object(RecyclebinPlugin, "read_bin_file", return_value=[mocked_file]):
        data = list(RecyclebinPlugin(target_win).read_recycle_bin(mocked_dir))

        assert data == [mocked_file] * 3


@pytest.mark.parametrize(
    ("version_number", "expected_header"),
    [
        (b"\x00" * 8, "header_v1"),
        (b"\x01" * 8, "header_v1"),
        (b"\x02" + b"\x00" * 7, "header_v2"),
    ],
)
def test_parse_header(target_win: Target, version_number: bytes, expected_header: str) -> None:
    recycle_plugin = RecyclebinPlugin(target_win)

    header = recycle_plugin.select_header(version_number)
    assert header is getattr(c_recyclebin, expected_header)


@pytest.mark.parametrize(
    "path",
    [
        "file/to/some/sid",
        "file/to/$recycle.bin/sid",
    ],
)
def test_read_bin_file_unknown(target_win: Target, path: str) -> None:
    recycle_plugin = RecyclebinPlugin(target_win)

    header_1 = c_recyclebin.header_v1(version=0, file_size=0x20, timestamp=0x20, filename="hello_world" + "\x00" * 249)

    with patch.object(Path, "open", mock_open(read_data=header_1.dumps())):
        normal_path = Path(path)

        output = next(recycle_plugin.read_bin_file(normal_path), None)

    assert output.filesize == 0x20
    assert output.path == "hello_world"


def test_reading_multiple_recyclebins(target_win: Target, recycle_bin: VirtualFilesystem) -> None:
    recycle_bin.map_file_fh("$ihello_world", io.BytesIO())

    target_win.fs.mount("C:\\$recycle.bin", recycle_bin)
    target_win.fs.mount("D:\\$recycle.bin", recycle_bin)
    target_win.add_plugin(RecyclebinPlugin)

    bin_file = Mock()
    with patch.object(RecyclebinPlugin, "read_bin_file", return_value=[bin_file]):
        recycle_bin_entries = list(target_win.recyclebin())
        assert recycle_bin_entries == [bin_file] * 2


def test_reading_deleted_directory_entries(target_win: Target, tmp_path: Path) -> None:
    """Test whether a records gets created for all the entries within a deleted directory."""
    filename = "c:\\path\\to\\hello_world\0"
    header = c_recyclebin.header_v2(version=2, file_size=0, filename_length=len(filename), filename=filename)

    tmp_path.joinpath("$Ihello_world").write_bytes(header.dumps())
    tmp_path.joinpath("$Rhello_world").mkdir()
    tmp_path.joinpath("$Rhello_world/data.txt").touch()

    recycle_bin = VirtualFilesystem()
    recycle_bin.map_dir("/", tmp_path)

    target_win.fs.mount("C:\\$recycle.bin", recycle_bin)
    target_win.add_plugin(RecyclebinPlugin)

    path = target_win.fs.path(filename.strip("\0"))

    records = list(target_win.recyclebin())
    start_dir = records[0]
    assert start_dir.path == filename.rstrip("\0")
    assert start_dir.deleted_path == "\\c:\\$recycle.bin\\$Rhello_world"
    assert start_dir.source == "\\c:\\$recycle.bin\\$Ihello_world"

    file_dir = records[1]
    assert file_dir.path == str(path.joinpath("data.txt"))
    assert file_dir.deleted_path == "\\c:\\$recycle.bin\\$Rhello_world\\data.txt"
    assert file_dir.source == "\\c:\\$recycle.bin\\$Ihello_world"


def test_recyclebin_wrong_prefix(target_win: Target, recycle_bin: VirtualFilesystem) -> None:
    recycle_bin.map_file_fh("hello_world", io.BytesIO())
    target_win.fs.mount("C:\\$recycle.bin", recycle_bin)
    target_win.add_plugin(RecyclebinPlugin)

    with patch.object(RecyclebinPlugin, "read_bin_file"):
        recycle_bin_entries = list(target_win.recyclebin())
        assert recycle_bin_entries == []


@pytest.mark.parametrize(
    ("path", "expected_output"),
    [
        ("C:/$Recycle.bin/sid-data/file", "sid-data"),
        ("$Recycle.bin/sid-data/file", "sid-data"),
        ("$Recycle.bin/just_another_file", "unknown"),
        ("C:/$Recycle.bin/just_another_file", "unknown"),
    ],
)
def test_find_sid_from_path(target_win: Target, path: str, expected_output: str) -> None:
    recycle_plugin = RecyclebinPlugin(target_win)

    assert recycle_plugin.find_sid(Path(path)) == expected_output



from __future__ import annotations

import sys
from typing import TYPE_CHECKING, BinaryIO

from dissect.cstruct import cstruct

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import export
from dissect.target.plugins.os.unix.locate.locate import BaseLocatePlugin

if TYPE_CHECKING:
    from collections.abc import Iterator

try:
    if sys.version_info >= (3, 14):
        from compression import zstd  # novermin
    else:
        from backports import zstd

    HAS_ZSTD = True
except ImportError:
    HAS_ZSTD = False

# Resource: https://git.sesse.net/?p=plocate @ db.h
plocate_def = """
#define MAGIC 0x00706c6f63617465             /* b'/x00plocate' */

struct header {
    uint32_t version;
    uint32_t hashtable_size;
    uint32_t extra_ht_slots;
    uint32_t num_docids;
    uint64_t hash_table_offset_bytes;
    uint64_t filename_index_offset_bytes;

    /* Version 1 and up only. */
    uint32_t max_version;   // Nominally 1 or 2, but can be increased if more features are added in a backward-compatible way.
    uint32_t zstd_dictionary_length_bytes;
    uint64_t zstd_dictionary_offset_bytes;

    /* Only if max_version >= 2, and only relevant for updatedb. */
    uint64_t directory_data_length_bytes;
    uint64_t directory_data_offset_bytes;
    uint64_t next_zstd_dictionary_length_bytes;
    uint64_t next_zstd_dictionary_offset_bytes;
    uint64_t conf_block_length_bytes;
    uint64_t conf_block_offset_bytes;

    // Only if max_version >= 2.
    uint8_t check_visibility;
    char padding[7];                         /* padding for alignment */
};
"""  # noqa : E501

PLocateRecord = TargetRecordDescriptor(
    "linux/locate/plocate",
    [
        ("path", "path"),
        ("path", "source"),
    ],
)

c_plocate = cstruct().load(plocate_def)


class PLocateFile:
    """Parser for plocate files.

    The ``plocate.db`` file contains a hashtable and trigrams to enable quick lookups of filenames.

    We've implemented a few methods to gather those for possible future use, but for the PLocatePlugin
    we're only interested in the filepaths stored in the database. Hence we don't use these methods.

    Roughly speaking, the plocate.db file has the following structure:
        - ``header`` (0x70 bytes)
        - zstd compressed ``filename``s (until start of ``filename_index_offset_bytes``),
          possibly including a dictionary
        - hashtables (offset and length in ``header``)
        - directory data (offset and length in ``header``)
        - possible zstd dictionary (offset and length in ``header``)
        - configuration block (offset and length in ``header``)

    No documentation other than the source code is available on the format of this file.

    References:
        - https://git.sesse.net/?p=plocate
    """

    HEADER_SIZE = 0x70  # 0x8 bytes magic + 0x68 bytes header
    NUM_OVERFLOW_SLOTS = 16
    TRIGRAM_SIZE_BYTES = 16

    def __init__(self, fh: BinaryIO):
        self.fh = fh

        magic = int.from_bytes(self.fh.read(8), byteorder="big")
        if magic != c_plocate.MAGIC:
            raise ValueError(f"Invalid plocate file magic. Expected b'/x00plocate', got {magic}")

        self.header = c_plocate.header(self.fh)
        self.zstd_dict = None

        if self.header.zstd_dictionary_offset_bytes:
            self.zstd_dict = zstd.ZstdDict(self.fh.read(self.header.zstd_dictionary_length_bytes))

        self.compressed_length_bytes = (
            self.header.filename_index_offset_bytes - self.HEADER_SIZE - self.header.zstd_dictionary_length_bytes
        )

    def __iter__(self) -> Iterator[str]:
        offsets = self.filename_index()
        for i in range(self.header.num_docids):
            self.fh.seek(offsets[i])
            block = zstd.decompress(self.fh.read(offsets[i + 1] - offsets[i]), zstd_dict=self.zstd_dict)
            for path in block.split(b"\x00"):
                yield path.decode(errors="surrogateescape")

    def filename_index(self) -> list[int]:
        """Return the filename index of the plocate.db file."""
        self.fh.seek(self.header.filename_index_offset_bytes)
        return c_plocate.uint64[self.header.num_docids + 1](self.fh)

    def hashtable(self) -> bytes:
        """Return the hashtable of the plocate.db file."""
        self.fh.seek(self.header.hash_table_offset_bytes)
        hashtable_size = (self.header.hashtable_size + self.NUM_OVERFLOW_SLOTS + 1) * self.TRIGRAM_SIZE_BYTES
        return self.fh.read(hashtable_size)


class PLocatePlugin(BaseLocatePlugin):
    """Unix plocate plugin."""

    __namespace__ = "plocate"

    path = "/var/lib/plocate/plocate.db"

    def check_compatible(self) -> None:
        if not self.target.fs.path(self.path).exists():
            raise UnsupportedPluginError(f"No plocate.db file found at {self.path}")

        if not HAS_ZSTD:
            raise UnsupportedPluginError(
                "Required dependency 'backports.zstd' is missing, install with `pip install backports.zstd` to use the PLocatePlugin"  # noqa: E501
            )

    @export(record=PLocateRecord)
    def locate(self) -> Iterator[PLocateRecord]:
        """Yield file and directory names from the plocate.db.

        ``plocate`` is the default package on Ubuntu 22 and newer to locate files.
        It replaces ``mlocate`` and GNU ``locate``.

        References:
            - https://manpages.debian.org/testing/plocate/plocate.1.en.html
            - https://git.sesse.net/?p=plocate
        """
        plocate = self.target.fs.path(self.path)
        plocate_file = PLocateFile(plocate.open())

        for path in plocate_file:
            yield PLocateRecord(
                path=self.target.fs.path(path),
                source=self.path,
            )



"""Filesystem and path related utilities."""

from __future__ import annotations

import fnmatch
import gzip
import hashlib
import io
import os
import re
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Any, BinaryIO, TextIO

from dissect.target.helpers.logging import get_logger

try:
    import lzma

    HAS_XZ = True
except ImportError:
    HAS_XZ = False

try:
    import bz2

    HAS_BZ2 = True
except ImportError:
    HAS_BZ2 = False

try:
    if sys.version_info >= (3, 14):
        from compression import zstd  # novermin
    else:
        from backports import zstd

    HAS_ZSTD = True
except ImportError:
    HAS_ZSTD = False

from dissect.target.exceptions import FileNotFoundError, SymlinkRecursionError
from dissect.target.helpers.polypath import (
    abspath,
    basename,
    commonpath,
    dirname,
    isabs,
    isreserved,
    join,
    normalize,
    normpath,
    relpath,
    split,
    splitdrive,
    splitext,
    splitroot,
)

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator, Sequence

    from typing_extensions import Self

    import dissect.target.filesystem as filesystem

if sys.version_info >= (3, 13):
    from dissect.target.helpers.compat.path_313 import PureDissectPath, TargetPath
elif sys.version_info >= (3, 12):
    from dissect.target.helpers.compat.path_312 import PureDissectPath, TargetPath
elif sys.version_info >= (3, 11):
    from dissect.target.helpers.compat.path_311 import PureDissectPath, TargetPath
elif sys.version_info >= (3, 10):  # noqa: UP036
    from dissect.target.helpers.compat.path_310 import PureDissectPath, TargetPath
else:
    raise RuntimeError("dissect.target requires at least Python 3.10")


log = get_logger(__name__)

re_glob_magic = re.compile(r"[*?[]")
re_glob_index = re.compile(r"(?<=\/)[^\/]*[*?[]")

__all__ = [
    "PureDissectPath",
    "TargetPath",
    "abspath",
    "basename",
    "commonpath",
    "dirname",
    "fs_attrs",
    "generate_addr",
    "glob_ext",
    "glob_split",
    "has_glob_magic",
    "isabs",
    "isreserved",
    "join",
    "normalize",
    "normpath",
    "open_decompress",
    "recurse",
    "relpath",
    "resolve_link",
    "reverse_read",
    "reverse_readlines",
    "split",
    "splitdrive",
    "splitext",
    "splitroot",
    "stat_result",
    "walk",
    "walk_ext",
]


def generate_addr(path: str | Path, alt_separator: str = "") -> int:
    if not alt_separator and isinstance(path, Path):
        alt_separator = (getattr(path, "parser", None) or path._flavour).altsep
    path = normalize(str(path), alt_separator=alt_separator)
    return int(hashlib.sha256(path.encode()).hexdigest()[:8], 16)


class stat_result:
    """Custom stat_result object, designed to mimick os.stat_result.

    The real stat_result is a CPython internal StructSeq, which kind of behaves like a namedtuple on steroids.
    We try to emulate some of that behaviour here.

    For consistency this class is also called stat_result.
    """

    __slots__ = {  # noqa: RUF023
        # Regular fields
        "st_mode": "protection bits",
        "st_ino": "inode",
        "st_dev": "device",
        "st_nlink": "number of hard links",
        "st_uid": "user ID of owner",
        "st_gid": "group ID of owner",
        "st_size": "total size, in bytes",
        # Regular timestamp fields (integer, float, nanosecond)
        "_st_atime": "integer time of last access",
        "_st_mtime": "integer time of last modification",
        "_st_ctime": "integer time of last change",
        "st_atime": "time of last access",
        "st_mtime": "time of last modification",
        "st_ctime": "time of last change",
        "st_atime_ns": "time of last access in nanoseconds",
        "st_mtime_ns": "time of last modification in nanoseconds",
        "st_ctime_ns": "time of last change in nanoseconds",
        # Extra optional fields
        "st_blksize": "blocksize for filesystem I/O",
        "st_blocks": "number of blocks allocated",
        "st_rdev": "device type (if inode device)",
        "st_flags": "user defined flags for file",
        "st_gen": "generation number",
        "st_birthtime": "time of creation",
        "st_file_attributes": "Windows file attribute bits",
        "st_fstype": "Type of filesystem",
        "st_reparse_tag": "Windows reparse tag",
        "st_birthtime_ns": "time of creation in nanoseconds",
        # Internal fields
        "_s": "internal tuple",
    }
    _field_count = len(__slots__) - 1

    def __init__(self, s: Sequence[Any]):
        if not isinstance(s, (list, tuple)) or not len(s) >= 10:
            raise TypeError(f"dissect.target.stat_result() takes an at least 10-sequence ({len(s)}-sequence given)")

        if len(s) > self._field_count:
            raise TypeError(
                f"dissect.target.stat_result() takes an at most {self._field_count}-sequence ({len(s)}-sequence given)"
            )

        s = tuple(s)
        if len(s) < self._field_count:
            s = s + ((None,) * (self._field_count - len(s)))

        self.st_mode = s[0]
        self.st_ino = s[1]
        self.st_dev = s[2]
        self.st_nlink = s[3]
        self.st_uid = s[4]
        self.st_gid = s[5]
        self.st_size = s[6]

        # Parse the values given to us in the "integer" field and calculate
        # the integer, float and nanosecond timestamps from there.
        self._st_atime, self.st_atime, self.st_atime_ns = self._parse_time(s[7])
        self._st_mtime, self.st_mtime, self.st_mtime_ns = self._parse_time(s[8])
        self._st_ctime, self.st_ctime, self.st_ctime_ns = self._parse_time(s[9])

        # Allow overriding the calculated values so a more accurate nanosecond timestamp can be specified
        self.st_atime = s[10] or self.st_atime
        self.st_mtime = s[11] or self.st_mtime
        self.st_ctime = s[12] or self.st_ctime
        self.st_atime_ns = s[13] or self.st_atime_ns
        self.st_mtime_ns = s[14] or self.st_mtime_ns
        self.st_ctime_ns = s[15] or self.st_ctime_ns

        self.st_blksize = s[16]
        self.st_blocks = s[17]
        self.st_rdev = s[18]
        self.st_flags = s[19]
        self.st_gen = s[20]
        self.st_birthtime = s[21]
        self.st_file_attributes = s[22]
        self.st_fstype = s[23]
        self.st_reparse_tag = s[24]
        self.st_birthtime_ns = s[25]

        # stat_result behaves like a tuple, but only with the first 10 fields
        # Note that this means it specifically uses the integer variants of the timestamps
        self._s = (
            self.st_mode,
            self.st_ino,
            self.st_dev,
            self.st_nlink,
            self.st_uid,
            self.st_gid,
            self.st_size,
            self._st_atime,
            self._st_mtime,
            self._st_ctime,
        )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, stat_result):
            other = other._s

        return self._s == other

    def __ne__(self, other: object) -> bool:
        return not self == other

    def __getitem__(self, item: int) -> int:
        return self._s[item]

    def __iter__(self) -> Iterator[int]:
        return iter(self._s)

    def __repr__(self) -> str:
        values = ", ".join(
            f"{k}={getattr(self, k)}" for k in self.__slots__ if k.startswith("st_") and getattr(self, k) is not None
        )
        return f"dissect.target.stat_result({values})"

    def _parse_time(self, ts: float) -> tuple[int, float, int]:
        ts_int = int(ts)
        ts_ns = int(ts * 1e9)

        return ts_int, ts_ns * 1e-9, ts_ns

    @classmethod
    def copy(cls, other: stat_result) -> Self:
        # First copy the basic 10 fields
        st = cls(list(other))
        # Then iterate and copy any other
        for attr in list(cls.__slots__.keys())[10 : cls._field_count]:
            try:
                setattr(st, attr, getattr(other, attr))
            except AttributeError:  # noqa: PERF203
                pass
        return st


def walk(
    path_entry: filesystem.FilesystemEntry,
    topdown: bool = True,
    onerror: Callable[[Exception], None] | None = None,
    followlinks: bool = False,
) -> Iterator[
    tuple[list[filesystem.FilesystemEntry], list[filesystem.FilesystemEntry], list[filesystem.FilesystemEntry]]
]:
    for path_list, dirs, files in walk_ext(path_entry, topdown, onerror, followlinks):
        dir_names = [d.name for d in dirs]
        file_names = [f.name for f in files]

        walk_path = join(path_entry.path, *[p.name for p in path_list[1:]])
        yield walk_path, dir_names, file_names

        if len(dir_names) != len(dirs):
            dirs[:] = [d for d in dirs if d.name in dir_names]


def walk_ext(
    path_entry: filesystem.FilesystemEntry,
    topdown: bool = True,
    onerror: Callable[[Exception], None] | None = None,
    followlinks: bool = False,
) -> Iterator[
    tuple[list[filesystem.FilesystemEntry], list[filesystem.FilesystemEntry], list[filesystem.FilesystemEntry]]
]:
    dirs: list[filesystem.FilesystemEntry] = []
    files: list[filesystem.FilesystemEntry] = []

    try:
        for entry in path_entry.scandir():
            if entry.is_dir():
                dirs.append(entry.get())
            else:
                files.append(entry.get())
    except Exception as e:
        if onerror is not None and callable(onerror):
            e.entry = path_entry
            onerror(e)
        return

    if topdown:
        yield [path_entry], dirs, files

    for direntry in dirs:
        if followlinks or not direntry.is_symlink():
            for xpath, xdirs, xfiles in walk_ext(direntry, topdown, onerror, followlinks):
                yield [path_entry, *xpath], xdirs, xfiles

    if not topdown:
        yield [path_entry], dirs, files


def recurse(entry: filesystem.FilesystemEntry) -> Iterator[filesystem.FilesystemEntry]:
    """Recursively walk the given :class:`FilesystemEntry`, yields :class:`DirEntry` instances."""
    yield entry

    if not entry.is_dir():
        return

    for direntry in entry.scandir():
        if direntry.is_dir(follow_symlinks=False):
            yield from recurse(direntry.get())
        else:
            yield direntry.get()


def glob_split(pattern: str, alt_separator: str = "") -> tuple[str, str]:
    """Split a pattern on path part boundaries on the first path part with a glob pattern.

    Args:
        pattern: A glob pattern to match names of filesystem entries against.
        alt_separator: An alternative path separator in use by the filesystem being matched.

    Returns:
        A tuple of a string with path parts up to the first path part that has a glob pattern and a string of
        the remaining path parts.
    """
    # re_glob_index expects a normalized pattern
    pattern = normalize(pattern, alt_separator=alt_separator)

    first_glob = re_glob_index.search(pattern)

    if not first_glob:
        return pattern, ""

    pos = first_glob.start()
    return pattern[:pos], pattern[pos:]


def glob_ext(direntry: filesystem.FilesystemEntry, pattern: str) -> Iterator[filesystem.FilesystemEntry]:
    """Recursively search and return filesystem entries matching a given glob pattern.

    Args:
        direntry: The filesystem entry relative to which to search.
        pattern: A glob pattern to match names of filesystem entries against.

    Yields:
        Matching filesystem entries (files and/or directories).
    """

    # Split the pattern on the last path part. base_name will contain the last path part (which is
    # '' if pattern ends with a /) and dir_name will contain the other parts.
    dir_name, base_name = split(pattern, alt_separator=direntry.fs.alt_separator)

    # The simple case where there are no globs.
    if not has_glob_magic(pattern):
        try:
            entry = direntry.get(pattern)
        except FileNotFoundError:
            pass
        else:
            # Patterns ending with a slash, so without a base_name, should match only directories.
            if base_name or entry.is_dir():
                yield entry
        return

    # The pattern has only one path part, so we can match directly against the files in direntry.
    if not dir_name:
        for entry in glob_ext1(direntry, base_name):
            yield entry
        return

    # If the pattern has more than one path part and these parts (dir_name) contain globs, we
    # recursively go over all the path parts and match them (glob_ext).
    # If these path parts have no globs, we get the path directly by name (glob_ext0).
    glob_in_dir = glob_ext if has_glob_magic(dir_name) else glob_ext0

    # If the pattern's last path part (base_name) has globs, we fnmatch it against the entries in
    # direntry (glob_ext1), otherwise we get the part directly by name (glob_ext0).
    glob_in_base = glob_ext1 if has_glob_magic(base_name) else glob_ext0

    for entry in glob_in_dir(direntry, dir_name):
        yield from glob_in_base(entry, base_name)


# These 2 helper functions non-recursively glob inside a literal directory.
def glob_ext1(direntry: filesystem.FilesystemEntry, pattern: str) -> Iterator[filesystem.FilesystemEntry]:
    """Match and return filesystem entries in a given filesystem entry based on pattern.

    Args:
        direntry: The filesystem entry relative to which to match the entries.
        pattern: A glob pattern to match names of filesystem entries against.

    Yields:
        Matching filesystem entries (files and/or directories).
    """
    if not direntry.is_dir():
        return

    entries = direntry.scandir()

    if pattern[0] != ".":
        # Do not return dot-files, unless they are explicitly searched for.
        entries = filter(lambda x: x.name[0] != ".", entries)

    for entry in entries:
        case_sensitive = entry.fs.case_sensitive
        name = entry.name if case_sensitive else entry.name.lower()
        pattern = pattern if case_sensitive else pattern.lower()
        if fnmatch.fnmatch(name, pattern):
            yield entry.get()


def glob_ext0(direntry: filesystem.FilesystemEntry, path: str) -> Iterator[filesystem.FilesystemEntry]:
    """Return the filesystem entry equal to the given path relative to direntry.

    Args:
        direntry: The filesystem entry relative to which to return the path.
        path: The path to the filesystem entry to return (if present). If path
              is an empty string (``''``) the ``direntry`` itself is returned.

    Yields:
        The matching filesystem entry (file or directory).
    """
    if path == "":
        # os.path.split() returns an empty path for paths ending with a directory separator.
        # E.g. 'q*x/' should match only directories.
        if direntry.is_dir():
            yield direntry
    elif direntry.is_dir():
        try:
            yield direntry.get(path)
        except FileNotFoundError:
            pass


def has_glob_magic(s: str) -> bool:
    return re_glob_magic.search(s) is not None


def resolve_link(
    fs: filesystem.Filesystem,
    link: str,
    path: str,
    *,
    alt_separator: str = "",
    previous_links: set[str] | None = None,
) -> filesystem.FilesystemEntry:
    """Resolves a symlink to its actual path.

    It stops resolving once it detects an infinite recursion loop.
    """

    link = normalize(link, alt_separator=alt_separator)
    path = normalize(path, alt_separator=alt_separator)

    # Create hash for entry based on path and link
    link_id = f"{path}{link}"
    hash_entry = hash(link_id)

    if not previous_links:
        previous_links = set()

    # Check whether the current entry was already resolved once.
    if hash_entry in previous_links:
        raise SymlinkRecursionError(f"Symlink loop detected for {link_id}")

    previous_links.add(hash_entry)

    if not isabs(link):
        cur_dirname = dirname(normpath(path))
        link = normpath(join(cur_dirname, link))

    # retrieve file from root
    entry = fs.get(link)

    if entry.is_symlink():
        entry = resolve_link(
            fs,
            entry.readlink(),
            link,
            alt_separator=entry.fs.alt_separator,
            previous_links=previous_links,
        )

    return entry


def open_decompress(
    path: TargetPath | None = None,
    mode: str = "rb",
    *,
    fileobj: BinaryIO | None = None,
    encoding: str | None = "UTF-8",
    errors: str | None = "backslashreplace",
    newline: str | None = None,
) -> BinaryIO | TextIO:
    """Open and decompress a file. Handles gz, bz2 and zstd files. Uncompressed files are opened as-is.

    When passing in an already opened ``fileobj``, the mode, encoding, errors and newline arguments are ignored.

    Args:
        path: The path to the file to open and decompress. It is assumed this path exists.
        mode: The mode in which to open the file.
        fileobj: The file-like object to open and decompress. This is mutually exclusive with path.
        encoding: The decoding for text streams. By default UTF-8 encoding is used.
        errors: The error handling for text streams. By default we're more lenient and use ``backslashreplace``.
        newline: How newlines are handled for text streams.

    Returns:
        An binary or text IO stream, depending on the mode with which the file was opened.

    Raises:
        ValueError: path and fileobj are mutually exclusive, but one of them is required

    Example:
        .. code-block:: python

            bytes_buf = open_decompress(Path("/dir/file.gz")).read()

            for line in open_decompress(Path("/dir/file.gz"), "rt"):
                print(line)
    """
    if path and fileobj:
        raise ValueError("path and fileobj are mutually exclusive")

    if not path and not fileobj:
        raise ValueError("path or fileobj is required")

    if path:
        file = path.open("rb")
    else:
        file = fileobj
        file.seek(0)

    magic = file.read(5)
    file.seek(0)

    if "b" in mode:
        # Reset the default encoding and errors mode in case of a binary stream
        encoding = None
        errors = None

    if magic[:2] == b"\x1f\x8b":
        return gzip.open(file, mode, encoding=encoding, errors=errors, newline=newline)

    if magic[:5] == b"\xfd7zXZ":
        if not HAS_XZ:
            raise RuntimeError("lzma compression detected, but missing optional python module")
        return lzma.open(file, mode, encoding=encoding, errors=errors, newline=newline)

    if magic[:3] == b"BZh" and 0x31 <= magic[3] <= 0x39:
        # In a valid bz2 header the 4th byte is in the range b'1' ... b'9'.
        if not HAS_BZ2:
            raise RuntimeError("bz2 compression detected, but missing optional python module")
        return bz2.open(file, mode, encoding=encoding, errors=errors, newline=newline)

    if magic[:4] in [b"\xfd\x2f\xb5\x28", b"\x28\xb5\x2f\xfd"]:
        if not HAS_ZSTD:
            raise RuntimeError("zstd compression detected, but missing optional python module")
        return zstd.open(file, mode=mode, encoding=encoding, errors=errors, newline=newline)

    if path:
        file.close()
        return path.open(mode, encoding=encoding, errors=errors, newline=newline)

    return file


def reverse_read(fh: BinaryIO, chunk_size: int = io.DEFAULT_BUFFER_SIZE, reverse_chunk: bool = True) -> Iterator[bytes]:
    """Like iterating over chunks of a binary file-like object, but starting from the end of the file.

    Args:
        fh: The file-like object (opened in binary mode) to read from.
        chunk_size: The chunk size to use for iterating over bytes (default: 8KB).
        reverse_chunk: Whether we should reverse the bytes of each chunk (default: True).

    Returns:
        An iterator of byte chunks, starting from the end of the file-like object and moving to the start.
    """

    offset = fh.seek(0, io.SEEK_END)

    while offset > 0:
        if offset < chunk_size:
            chunk_size = offset
        offset -= chunk_size
        fh.seek(offset)
        buf = fh.read(chunk_size)
        yield bytes(reversed(buf)) if reverse_chunk else buf


def reverse_readlines(fh: TextIO, chunk_size: int = 1024 * 1024 * 8) -> Iterator[str]:
    """Like iterating over a ``TextIO`` file-like object, but starting from the end of the file.

    Args:
        fh: The file-like object (opened in text mode) to iterate lines from.
        chunk_size: The chunk size to use for iterating over lines.

    Returns:
        An iterator of lines from the file-like object, in reverse.
    """
    offset = fh.seek(0, io.SEEK_END) & ((1 << 64) - 1)
    lines = []

    prev_offset = offset
    while offset > 0:
        if offset < chunk_size:
            chunk_size = offset
        offset -= chunk_size
        fh.seek(offset)

        lines = []
        # tell() on TextIO returns a cookie which includes encode/decoder state
        # Lower 64 bit are the file position
        # https://peps.python.org/pep-3116/#text-i-o
        # See TextIOWrapper._pack_cookie in _pyio.py for more detail
        while (fh.tell() & ((1 << 64) - 1)) < prev_offset:
            try:
                lines.append(fh.readline())
            except UnicodeDecodeError:  # noqa: PERF203
                offset += 1
                fh.seek(offset)

        yield from reversed(lines[1:])

        if prev_offset == offset:
            # Previous lines are unreadable due to decoding errors
            raise UnicodeDecodeError(fh.encoding, b"", 0, offset + 1, "failed to decode line")

        prev_offset = offset

    if lines:
        yield lines[0]


def fs_attrs(
    path: os.PathLike | str | bytes,
    follow_symlinks: bool = True,
) -> dict[os.PathLike | str | bytes, bytes]:
    """Return the extended attributes for a given path on the local filesystem.

    This is currently only implemented for Linux using os.listxattr and related functions.

    Args:
        path: The path to get the extended attributes for.
        follow_symlinks: Wether to follow the symlink if the given path is a symlink.

    Returns:
        A dict containing the attribute names as keys and their values.
    """
    attrs = {}
    if hasattr(os, "listxattr"):
        # os.listxattr etc. are only available on Linux
        attr_names = os.listxattr(path, follow_symlinks=follow_symlinks)
        for attr_name in attr_names:
            attrs[attr_name] = os.getxattr(path, attr_name, follow_symlinks=follow_symlinks)

    return attrs



from __future__ import annotations

from typing import TYPE_CHECKING

try:
    from ruamel.yaml import YAML
except ImportError:
    raise ImportError("Missing ruamel.yaml dependency")

from dissect.target import container
from dissect.target.loader import Loader

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.target import Target


class TargetLoader(Loader):
    """Load target files."""

    def __init__(self, path: Path, **kwargs):
        super().__init__(path, **kwargs)
        self.definition = YAML(typ="safe").load(path.open("rb"))

    @staticmethod
    def detect(path: Path) -> bool:
        return path.suffix.lower() == ".target"

    def map(self, target: Target) -> None:
        for disk in self.definition["disks"]:
            target.disks.add(container.open(disk))



from __future__ import annotations

import plistlib
from typing import TYPE_CHECKING

from flow.record.fieldtypes import posix_path

from dissect.target.filesystem import Filesystem
from dissect.target.helpers.record import MacOSUserRecord
from dissect.target.plugin import OperatingSystem, export
from dissect.target.plugins.os.unix.bsd.darwin._os import (
    DarwinPlugin,
    detect_macho_arch,
)
from dissect.target.target import Target

if TYPE_CHECKING:
    from collections.abc import Iterator

    from typing_extensions import Self

    from dissect.target.filesystem import Filesystem
    from dissect.target.target import Target


class MacOSPlugin(DarwinPlugin):
    VERSION = "/System/Library/CoreServices/SystemVersion.plist"
    GLOBAL = "/Library/Preferences/.GlobalPreferences.plist"
    SYSTEM = "/Library/Preferences/SystemConfiguration/preferences.plist"

    def __init__(self, target: Target):
        super().__init__(target)
        self._add_mounts()

    @classmethod
    def detect(cls, target: Target) -> Filesystem | None:
        for fs in target.filesystems:
            if fs.exists("/Library") and fs.exists("/Applications") and not fs.exists("/private/var/mobile"):
                return fs

        return None

    @classmethod
    def create(cls, target: Target, sysvol: Filesystem) -> Self:
        target.fs.mount("/", sysvol)
        return cls(target)

    def _add_mounts(self) -> None:
        data_fs = None
        for fs in self.target.filesystems:
            if fs.__type__ == "apfs" and fs.apfs.role.name == "DATA":
                data_fs = fs
                break

        if data_fs and (path := self.target.fs.path("/usr/share/firmlinks")).exists():
            firmlinks = [line.strip().split("\t") for line in path.open("rt") if line.strip()]
            layer = self.target.fs.append_layer()

            for dst, src in firmlinks:
                if data_fs.exists(src):
                    layer.map_file_entry(dst, data_fs.get(src))

        # TODO: synthetic.conf

    @export(property=True)
    def hostname(self) -> str | None:
        try:
            preferences = plistlib.load(self.target.fs.path(self.SYSTEM).open())
            return preferences["System"]["Network"]["HostNames"]["LocalHostName"]

        except FileNotFoundError:
            pass

    @export(property=True)
    def ips(self) -> list[str] | None:
        return list(set(map(str, self.target.network.ips())))

    @export(property=True)
    def version(self) -> str | None:
        try:
            systemVersion = plistlib.load(self.target.fs.path(self.VERSION).open())
        except FileNotFoundError:
            pass
        else:
            productName = systemVersion["ProductName"]
            productUserVisibleVersion = systemVersion["ProductUserVisibleVersion"]
            productBuildVersion = systemVersion["ProductBuildVersion"]
            return f"{productName} {productUserVisibleVersion} ({productBuildVersion})"

    @export(record=MacOSUserRecord)
    def users(self) -> Iterator[MacOSUserRecord]:
        try:
            for path in self.target.fs.path("/var/db/dslocal/nodes/Default/users/").glob("*.plist"):
                user = plistlib.load(path.open())

                # The home directory of a user account can be null,
                # but a user account can also have multiply home directories e.g. the root account.
                # https://developer.apple.com/documentation/foundation/filemanager/1642853-homedirectory/
                for home_dir in user.get("home", [None]):
                    yield MacOSUserRecord(
                        name=user.get("name", [None])[0],
                        passwd=user.get("passwd", [None])[0],
                        uid=user.get("uid", [None])[0],
                        gid=user.get("gid", [None])[0],
                        gecos=user.get("realname", [None])[0],
                        home=posix_path(home_dir) if home_dir else None,
                        shell=user.get("shell", [None])[0],
                        source=path,
                    )
        except FileNotFoundError:
            pass

    @export(property=True)
    def os(self) -> str:
        return OperatingSystem.MACOS.value

    @export(property=True)
    def architecture(self) -> str | None:
        if arch := detect_macho_arch(
            paths=[
                "/bin/bash",
                "/bin/sh",
                "/bin/cp",
                "/bin/ls",
                "/bin/ps",
            ],
            fs=self.target.fs,
        ):
            return f"{arch}-macos"
        return None



from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING
from unittest.mock import patch

import pytest
from dissect.ntfs import NTFS

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.filesystems.dir import DirectoryFilesystem
from dissect.target.helpers import loaderutil
from dissect.target.tools.mount import main as target_mount
from dissect.target.volume import Volume
from tests._utils import absolute_path

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.target import Target


@pytest.fixture
def mock_ntfs_dirfs(tmp_path: Path) -> Path:
    root = tmp_path / "mock_ntfs"
    root.mkdir()

    # Create files
    (root / "$Boot").touch()
    (root / "$Extend_$Usnjrnl_$J").touch()
    (root / "$Secure_$SDS").touch()

    # Only need this to exist up until the root directory record to make dissect.ntfs happy
    with absolute_path("_data/plugins/filesystem/ntfs/mft/mft.raw").open("rb") as fh:
        (root / "$MFT").write_bytes(fh.read(10 * 1024))

    return root


def test_duplicate_volume_name(target_bare: Target, monkeypatch: pytest.MonkeyPatch) -> None:
    with monkeypatch.context() as m:
        m.setattr("sys.argv", ["target-mount", "mock-target", "mock-mount"])
        m.setattr("dissect.target.tools.mount.HAS_FUSE", True)

        with (
            patch("dissect.target.tools.mount.open_target", return_value=target_bare),
            patch("dissect.target.tools.mount.FUSE", create=True) as MockFUSE,
            patch("dissect.target.tools.mount.DissectMount", create=True) as MockDissectMount,
        ):
            target_bare.volumes.add(Volume(BytesIO(), 1, 0, 0, None, name="first"))
            target_bare.volumes.add(Volume(BytesIO(), 2, 0, 0, None, name="second"))
            target_bare.volumes.add(Volume(BytesIO(), 3, 0, 0, None, name="second_1"))
            target_bare.volumes.add(Volume(BytesIO(), 4, 0, 0, None, name="second"))

            mock_fs = VirtualFilesystem()
            mock_fs.volume = target_bare.volumes[1]
            target_bare.filesystems.add(mock_fs)
            mock_fs = VirtualFilesystem()
            mock_fs.volume = target_bare.volumes[2]
            target_bare.filesystems.add(mock_fs)
            mock_fs = VirtualFilesystem()
            mock_fs.volume = target_bare.volumes[3]
            target_bare.filesystems.add(mock_fs)

            target_mount()

            MockFUSE.assert_called_once()
            MockDissectMount.assert_called_once()
            vfs = MockDissectMount.call_args[0][0]

            assert vfs.listdir("/volumes") == ["first", "second", "second_1", "second_2"]
            assert vfs.listdir("/filesystems") == ["second", "second_1", "second_2"]


def test_mounting_multi_volume_filesystem(target_bare: Target, monkeypatch: pytest.MonkeyPatch) -> None:
    with monkeypatch.context() as m:
        m.setattr("sys.argv", ["target-mount", "mock-target", "mock-mount"])
        m.setattr("dissect.target.tools.mount.HAS_FUSE", True)

        with (
            patch("dissect.target.tools.mount.open_target", return_value=target_bare),
            patch("dissect.target.tools.mount.FUSE", create=True) as MockFUSE,
            patch("dissect.target.tools.mount.DissectMount", create=True) as MockDissectMount,
        ):
            volumes = [
                Volume(BytesIO(), 1, 0, 0, None, name="first"),
                Volume(BytesIO(), 2, 0, 0, None, name="second"),
            ]

            for vol in volumes:
                target_bare.volumes.add(vol)

            mock_fs = VirtualFilesystem()
            mock_fs.volume = volumes
            target_bare.filesystems.add(mock_fs)

            target_mount()

            MockFUSE.assert_called_once()
            MockDissectMount.assert_called_once()
            vfs = MockDissectMount.call_args[0][0]

            assert vfs.listdir("/filesystems") == ["first", "second"]


def test_mounting_virtual_ntfs_filesystem(
    target_bare: Target,
    monkeypatch: pytest.MonkeyPatch,
    mock_ntfs_dirfs: Path,
) -> None:
    with monkeypatch.context() as m:
        m.setattr("sys.argv", ["target-mount", "mock-target", "mock-mount"])
        m.setattr("dissect.target.tools.mount.HAS_FUSE", True)

        with (
            patch("dissect.target.tools.mount.open_target", return_value=target_bare),
            patch("dissect.target.tools.mount.FUSE", create=True) as MockFUSE,
            patch("dissect.target.tools.mount.DissectMount", create=True) as MockDissectMount,
        ):
            # Use the mock_ntfs_dirfs fixture for the NTFS files
            dir_fs = DirectoryFilesystem(mock_ntfs_dirfs)
            virtual_fs = VirtualFilesystem()
            virtual_fs.map_fs("", dir_fs)

            # Add the VirtualFileSystem
            target_bare.filesystems.add(virtual_fs)
            # Mount the VirtualFileSystem as c:
            target_bare.fs.mount("c:", virtual_fs)

            loaderutil.add_virtual_ntfs_filesystem(
                target_bare,
                virtual_fs,
                usnjrnl_path="$Extend_$Usnjrnl_$J",
                sds_path="$Secure_$SDS",
            )

            # The loaderutil should now have made a mock NTFS
            ntfs_obj = getattr(virtual_fs, "ntfs", None)
            assert isinstance(ntfs_obj, NTFS)

            target_mount()

            MockFUSE.assert_called_once()
            MockDissectMount.assert_called_once()
            vfs = MockDissectMount.call_args[0][0]

            # Only the VirtualFileSystem should be available in /filesystems, not the mock NTFS
            assert vfs.listdir("/filesystems") == ["fs_0"]
            # The VirtualFileSystem should have the NTFS files
            required_ntfs_files = ["$Extend_$Usnjrnl_$J", "$MFT", "$Secure_$SDS", "$Boot"]
            assert all(file in required_ntfs_files for file in vfs.listdir("/filesystems/fs_0"))
            # The c: volume should be mounted, and should be the only mount
            assert vfs.listdir("/fs") == ["c:"]



from __future__ import annotations

import struct
import subprocess
from configparser import ConfigParser
from configparser import Error as ConfigParserError
from io import BytesIO
from typing import TYPE_CHECKING, Any, BinaryIO, TextIO

from dissect.util.hash.jenkins import lookup8

from dissect.target.filesystems.nfs import NfsFilesystem
from dissect.target.filesystems.vmtar import VmtarFilesystem
from dissect.target.helpers.sunrpc import client
from dissect.target.helpers.sunrpc.client import LocalPortPolicy

try:
    from dissect.hypervisor.util.envelope import (
        HAS_PYCRYPTODOME,
        HAS_PYSTANDALONE,
        Envelope,
        KeyStore,
    )

    HAS_ENVELOPE = HAS_PYCRYPTODOME or HAS_PYSTANDALONE
except ImportError:
    HAS_ENVELOPE = False

from dissect.target.filesystems import tar
from dissect.target.plugin import OperatingSystem, export
from dissect.target.plugins.os.unix._os import UnixPlugin

if TYPE_CHECKING:
    from typing_extensions import Self

    from dissect.target.filesystem import Filesystem, VirtualFilesystem
    from dissect.target.helpers.fsutil import TargetPath
    from dissect.target.target import Target


class ESXiPlugin(UnixPlugin):
    """ESXi OS plugin

    ESXi partitioning varies between versions. Generally, specific partition numbers have special meaning.

    The following is a list of known partition numbers:
        1: EFI boot
        5: BOOTBANK1
        6: BOOTBANK2
        7: vmkcore (ESXi 6), OSDATA / LOCKER (ESXi 7)
        8: store (ESXi 6), HDD VMFS datastore (ESXi 7)
        9: vmkcore (ESXi 6)
    """

    def __init__(self, target: Target):
        super().__init__(target)
        self._mount_nfs_shares()

    @classmethod
    def detect(cls, target: Target) -> Filesystem | None:
        # First handle 'simple' case where we have to deal with a live collection
        for fs in target.filesystems:
            if fs.path("/etc/vmware/esx.conf").exists():
                return fs

        bootbanks = [
            fs for fs in target.filesystems if fs.path("boot.cfg").exists() and list(fs.path("/").glob("*.v00"))
        ]

        cfgs = [(fs, parse_boot_cfg(fs.path("boot.cfg").open("rt"))) for fs in bootbanks]
        cfgs.sort(key=lambda pair: pair[1].get("updated", 0), reverse=True)

        return cfgs[0][0] if cfgs else None

    @classmethod
    def create(cls, target: Target, sysvol: Filesystem) -> Self:
        if sysvol.path("/etc/vmware/esx.conf").exists():
            target.fs.mount("/", sysvol)
            return cls(target)

        cfg = parse_boot_cfg(sysvol.path("boot.cfg").open("rt"))

        # Mount all the visor tars in individual filesystem layers
        _mount_modules(target, sysvol, cfg)

        # Create a root layer for the "local state" filesystem
        # This stores persistent configuration data
        local_layer = target.fs.append_layer()

        # Mount the local.tgz to the local state layer
        _mount_local(target, local_layer)

        # Mount all other filesystems (VMFS, FAT16)
        _mount_filesystems(target, sysvol, cfg)

        obj = cls(target)

        # Symlink the /var/log directory to the correct destination (if available)
        _link_log_dir(target, cfg, obj)

        return obj

    @export(property=True)
    def hostname(self) -> str:
        if hostname := self.target.esxconf.get("/adv/Misc/HostName"):
            return hostname.split(".", 1)[0]
        return "localhost"

    @export(property=True)
    def domain(self) -> str | None:
        if hostname := self.target.esxconf.get("/adv/Misc/HostName"):
            return hostname.partition(".")[2]
        return None

    @export(property=True)
    def ips(self) -> list[str]:
        result = set()
        host_ip = self.target.esxconf.get("/adv/Misc/HostIPAddr")
        mgmt_ip = self.target.esxconf.get("/adv/Net/ManagementAddr")

        if host_ip:
            result.add(host_ip)
        if mgmt_ip:
            result.add(mgmt_ip)

        return list(result)

    @export(property=True)
    def version(self) -> str | None:
        boot_cfg = self.target.fs.path("/bootbank/boot.cfg")
        if not boot_cfg.exists():
            # Default to retrieve version, but without build number
            return self.target.esxconf.get("/resourceGroups/version")

        for line in boot_cfg.read_text().splitlines():
            if not line.startswith("build="):
                continue

            _, _, version = line.partition("=")
            return f"VMware ESXi {version.strip()}"
        return None

    @export(property=True)
    def os(self) -> str:
        return OperatingSystem.ESXI.value

    def _mount_nfs_shares(self) -> None:
        """Mount NFS shares found in the configstore."""
        if not self.target.has_function("configstore.get"):
            self.target.log.warning("No configstore found, unable to mount NFS shares")
            return

        nfs_shares: dict[str, Any] = (
            self.target.configstore.get("esx", {}).get("storage", {}).get("nfs_v3_datastores", {})
        )
        if not nfs_shares:
            if self._is_nfs_enabled:
                self.target.log.info("No NFS shares found in datastore")
            return

        for key, nfs_share in nfs_shares.items():
            # Parse the NFS share configuration
            user_value: dict[str, Any] = nfs_share.get("user_value", {})
            nfs_ip = user_value.get("hostname", "")
            volume_name = user_value.get("volume_name", "")
            remote_share = user_value.get("remote_share", "")
            if not nfs_ip or not volume_name or not remote_share:
                self.target.log.warning("Invalid NFS share configuration with key: %s", key)
                continue
            mount_point = f"/vmfs/volumes/{volume_name}"
            self._add_nfs(nfs_ip, remote_share, mount_point)

    def _add_nfs(self, nfs_ip: str, remote_share: str, mount_alias: str) -> None:
        """Mount NFS share to the target."""

        if not self._is_nfs_enabled:
            self._log_nfs_mount_disabled(nfs_ip, remote_share, mount_alias)
            return

        uuid = nfs_volume_uuid(nfs_ip, remote_share)
        mount_point = f"/vmfs/volumes/{uuid}"
        try:
            self.target.log.debug("Mounting NFS share %s at %s with alias %s", remote_share, mount_point, mount_alias)

            # On ESXi, there is typically only a single root user.
            # Besides, UnixPlugin::users does not work (see issue https://github.com/fox-it/dissect.target/issues/1093).
            # Moreover, socket rebinding is not implemented on ESXi.
            # Therefore, we try logging only as root. This implies that he NFS share has `no_root_squash` enabled.
            # According to the docs, ESxi mounts NFS shares using root privileges (https://www.vmware.com/docs/vmw-best-practices-running-nfs-vmware-vsphere)  # noqa: E501
            credentials = client.auth_unix("machine", 0, 0, [])
            nfs = NfsFilesystem.connect(nfs_ip, remote_share, credentials, LocalPortPolicy.PRIVILEGED)

            self.target.fs.mount(mount_point, nfs)
            self.target.fs.symlink(mount_point, mount_alias)
        except Exception as e:
            self.target.log.warning(
                "Failed to mount NFS share %s:%s at %s",
                nfs_ip,
                remote_share,
                mount_alias,
            )
            self.target.log.debug("", exc_info=e)


def _mount_modules(target: Target, sysvol: Filesystem, cfg: dict[str, str]) -> None:
    modules = [m.strip() for m in cfg["modules"].split("---")]

    for module in modules:
        module_path = sysvol.path(module)
        if not module_path.exists():
            target.log.warning("Non existent module: %s", module)
            continue

        tfs = None
        if module_path.name.endswith((".tar.gz", ".tgz")):
            tfs = tar.TarFilesystem(module_path.open())
        elif module_path.suffix.startswith(".v"):
            try:
                tfs = VmtarFilesystem(module_path.open())
            except Exception as e:
                target.log.warning("%s, skipping file %s", str(e), module_path)
        if tfs:
            target.fs.append_layer().mount("/", tfs)


def _mount_local(target: Target, local_layer: VirtualFilesystem) -> None:
    local_tgz = target.fs.path("local.tgz")
    local_tgz_ve = target.fs.path("local.tgz.ve")
    local_fs = None

    if local_tgz.exists():
        local_fs = tar.TarFilesystem(local_tgz.open())
    elif local_tgz_ve.exists():
        # In the case "encryption.info" does not exist, but ".#encryption.info" does
        encryption_info = next(target.fs.path("/").glob("*encryption.info"), None)
        if not local_tgz_ve.exists() or not encryption_info.exists():
            raise ValueError("Unable to find valid configuration archive")

        local_fs = _create_local_fs(target, local_tgz_ve, encryption_info)
    else:
        target.log.warning("No local.tgz or local.tgz.ve found, skipping local state")

    if local_fs:
        local_layer.mount("/", local_fs)


def _decrypt_envelope(local_tgz_ve: TargetPath, encryption_info: TargetPath) -> BinaryIO:
    """Decrypt ``local.tgz.ve`` ourselves with hard-coded keys."""
    envelope = Envelope(local_tgz_ve.open())
    keystore = KeyStore.from_text(encryption_info.read_text("utf-8"))
    return BytesIO(envelope.decrypt(keystore.key, aad=b"ESXConfiguration"))


def _decrypt_crypto_util(local_tgz_ve: TargetPath) -> BytesIO | None:
    """Decrypt ``local.tgz.ve`` using ESXi ``crypto-util``.

    We write to stdout, but this results in ``crypto-util`` exiting with a non-zero return code
    and stderr containing an I/O error message. The file does get properly decrypted, so we return
    ``None`` if there are no bytes in stdout which would indicate it actually failed.
    """

    result = subprocess.run(
        [
            "crypto-util",
            "envelope",
            "extract",
            "--aad",
            "ESXConfiguration",
            f"/{local_tgz_ve.as_posix()}",
            "-",
        ],
        capture_output=True,
    )

    if len(result.stdout) == 0:
        return None

    return BytesIO(result.stdout)


def _create_local_fs(target: Target, local_tgz_ve: TargetPath, encryption_info: TargetPath) -> tar.TarFilesystem | None:
    local_tgz = None

    if HAS_ENVELOPE:
        try:
            local_tgz = _decrypt_envelope(local_tgz_ve, encryption_info)
        except NotImplementedError:
            target.log.debug("Failed to decrypt %s, likely TPM encrypted", local_tgz_ve)
    else:
        target.log.debug("Skipping static decryption because of missing crypto module")

    if not local_tgz and target.name == "local":
        target.log.info(
            "local.tgz is encrypted but static decryption failed, attempting dynamic decryption using crypto-util"
        )
        local_tgz = _decrypt_crypto_util(local_tgz_ve)

        if local_tgz is None:
            target.log.warning("Dynamic decryption of %s failed", local_tgz_ve)
    else:
        target.log.warning("local.tgz is encrypted but static decryption failed and no dynamic decryption available!")

    if local_tgz:
        return tar.TarFilesystem(local_tgz)
    return None


def _mount_filesystems(target: Target, sysvol: Filesystem, cfg: dict[str, str]) -> None:
    version = cfg["build"]

    osdata_fs = None
    locker_fs = None
    for fs in target.filesystems:
        if fs.__type__ == "fat":
            fs.volume.seek(512)
            magic, uuid1, uuid2, uuid3, uuid4 = struct.unpack("<16sIIH6s", fs.volume.read(32))
            if magic != b"VMWARE FAT16    ":
                continue

            fs_uuid = f"{uuid1:08x}-{uuid2:08x}-{uuid3:04x}-{uuid4.hex()}"
            target.fs.mount(f"/vmfs/volumes/{fs_uuid}", fs)
            if not fs.volume.name.startswith("part_"):  # dissect.target quirk
                target.fs.symlink(f"/vmfs/volumes/{fs_uuid}", f"/vmfs/volumes/{fs.volume.name}")

            # ESXi 7 relies on volume names
            # ESXi 6 uses volume numbers
            if fs.volume.name in ("BOOTBANK1", "BOOTBANK2") or (
                fs.volume.number
                in (
                    5,
                    6,
                )
                and fs.exists("boot.cfg")
            ):
                if fs is sysvol:
                    target.fs.symlink(f"/vmfs/volumes/{fs_uuid}", "/bootbank")
                else:
                    target.fs.symlink(f"/vmfs/volumes/{fs_uuid}", "/altbootbank")

            # /store == partition number 8
            if version and version[0] == "6" and fs.volume.number == 8:
                target.fs.symlink(f"/vmfs/volumes/{fs_uuid}", "/store")
                target.fs.symlink("/store", "/locker")

        elif fs.__type__ == "vmfs":
            target.fs.mount(f"/vmfs/volumes/{fs.vmfs.uuid}", fs)
            target.fs.symlink(f"/vmfs/volumes/{fs.vmfs.uuid}", f"/vmfs/volumes/{fs.vmfs.label}")

            if fs.volume.name in ("OSDATA", "LOCKER"):
                target.fs.symlink(
                    f"/vmfs/volumes/{fs.vmfs.uuid}",
                    f"/vmfs/volumes/{fs.volume.name}-{fs.vmfs.uuid}",
                )

                if fs.volume.name == "OSDATA":
                    osdata_fs = fs
                elif fs.volume.name == "LOCKER":
                    locker_fs = fs

    # Symlink /scratch from locker.conf
    # A path _should_ always be set here, even in default installs where it points to OSDATA
    # This file is confusingly called locker.conf
    locker_conf = target.fs.path("/etc/vmware/locker.conf")
    if locker_conf.exists():
        scratch_path = locker_conf.read_text().strip().partition(" ")[0]
        if scratch_path:
            target.fs.symlink(scratch_path, "/scratch")
        else:
            target.log.warning("Scratch path from locker.conf is empty?")
    else:
        target.log.warning("No locker.conf!")

    # Symlink /locker
    # This is marked as legacy in ESXi 7, so may need changing when ESXi 8 comes out
    # Apparently the order is: OSDATA partition -> LOCKER partition -> RAM disk
    # We can't support RAM disk so only check OSDATA and LOCKER
    if osdata_fs:
        target.fs.symlink(f"/vmfs/volumes/OSDATA-{osdata_fs.vmfs.uuid}", "/var/lib/vmware/osdata")
        target.fs.symlink("/var/lib/vmware/osdata/store", "/store")
        target.fs.symlink("/var/lib/vmware/osdata/locker", "/locker")

    elif locker_fs:
        target.fs.symlink(f"/vmfs/volumes/LOCKER-{locker_fs.vmfs.uuid}", "/locker")


def _link_log_dir(target: Target, cfg: dict[str, str], plugin_obj: ESXiPlugin) -> None:
    version = cfg["build"]

    # Don't really know how ESXi does this, but let's just take a shortcut for now
    log_dir = None
    if version and version[0] == "7":
        try:
            log_dir = target.configstore._configstore["esx"]["syslog"]["global_settings"][""]["user_value"]["log_dir"]
        except KeyError:
            target.log.warning("Failed to read log_dir from configstore, falling back to /scratch/log")
            log_dir = "/scratch/log"
    elif version and version[0] == "6":
        vmsyslog_file = target.fs.path("/etc/vmsyslog.conf")
        if vmsyslog_file.exists():
            try:
                vmsyslog_conf = ConfigParser()
                vmsyslog_conf.read_string(vmsyslog_file.read_text())

                log_dir = vmsyslog_conf.get("vmsyslog", "logdir")
                if log_dir == "<none>":
                    log_dir = "/scratch/log"
            except ConfigParserError as e:
                target.log.warning(
                    "Failed to read log_dir from vmsyslog.conf, falling back to /scratch/log",
                    exc_info=e,
                )
                log_dir = "/scratch/log"

    if log_dir:
        target.fs.symlink(log_dir, "/var/log")


def parse_boot_cfg(fh: TextIO) -> dict[str, str]:
    cfg = {}
    for line in fh:
        line = line.strip()
        if not line or line.startswith("#"):
            continue

        key, _, value = line.partition("=")

        cfg[key.strip()] = value.strip()

    return cfg


def nfs_volume_uuid(host: str, path: str) -> str:
    """Generate a UUID for an NFS volume based on the host and path.

    This is used to create a unique identifier for NFS volumes in ESXi.
    """

    h1 = lookup8(host.encode(), 42)  # 42 is starting value
    h2 = lookup8(path.encode(), h1)

    low, high = h2 & 0xFFFFFFFF, ((h2 >> 32) & 0xFFFFFFFF)
    return f"{low:8x}-{high:8x}"



from __future__ import annotations

import gzip

from dissect.target.containers.raw import RawContainer
from dissect.target.target import Target
from dissect.target.volumes.md import MdVolumeSystem
from tests._utils import absolute_path


def test_md() -> None:
    with gzip.open(absolute_path("_data/volumes/md/md-nested.bin.gz"), "rb") as fh:
        assert MdVolumeSystem.detect_volume(fh)

        sets = list(MdVolumeSystem.open_all([fh]))
        assert len(sets) == 1

        md = sets[0]
        assert len(md.volumes) == 1
        assert md.volumes[0].size == md.md.configurations[0].virtual_disks[0].size


def test_nested_md_lvm() -> None:
    with gzip.open(absolute_path("_data/volumes/md/md-nested.bin.gz"), "rb") as fh:
        container = RawContainer(fh)

        t = Target()
        t.disks.add(container)
        t.apply()

        assert len(t.filesystems) == 1
        assert t.fs.path("file.txt").read_text() == "wow it worked\n"



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import PluginError, UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.helpers.regutil import RegistryKey

RegistryKeyRecord = TargetRecordDescriptor(
    "windows/registry/regf/key",
    [
        ("datetime", "ts"),
        ("string", "path"),
        ("string", "key"),
        ("string", "source"),
    ],
)


RegistryValueRecord = TargetRecordDescriptor(
    "windows/registry/regf/value",
    [
        ("datetime", "ts"),
        ("string", "path"),
        ("string", "key"),
        ("string", "name"),
        ("dynamic", "value"),  # flow loophole
        ("varint", "data_type"),
        ("string", "source"),
    ],
)


class RegfPlugin(Plugin):
    """Regf dump plugin."""

    def check_compatible(self) -> None:
        try:
            self.target.registry.root()
        except PluginError:
            raise UnsupportedPluginError("Registry plugin not loaded")

    @export(record=[RegistryKeyRecord, RegistryValueRecord])
    def regf(self) -> Iterator[RegistryKeyRecord | RegistryValueRecord]:
        """Return all registry keys and values.

        The Windows Registry is a hierarchical database that stores low-level settings for the Windows operating system
        and for applications that opt to use it.

        References:
            - https://en.wikipedia.org/wiki/Windows_Registry

        Yields RegistryKeyRecords and RegistryValueRecords

        RegistryKeyRecord fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            ts (datetime): The registry key last modified time.
            path (string): The key path.
            key (string): The key name.
            source (string): The hive file path.

        RegistryValueRecord fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            ts (datetime): The registry key last modified time.
            path (string): The key path.
            key (string): The key name.
            name (string): The value name.
            value (string): The value.
            source (string): The hive file path.
        """
        self.target.registry.load_user_hives()
        mappings = self.target.registry.mappings()

        for name, hive, path in self.target.registry.iterhives():
            if name in mappings:
                name = mappings[name]
            elif name.startswith("S-"):
                # Bit of a nasty hack
                name = f"HKEY_USERS\\{name}"

            yield from self.walk(hive.root(), name, path)

    def walk(self, key: RegistryKey, parent: str, path: str) -> Iterator[RegistryKeyRecord | RegistryValueRecord]:
        yield RegistryKeyRecord(
            ts=key.timestamp,
            path=parent,
            key=key.name,
            source=path,
            _target=self.target,
        )

        for value in key.values():
            yield RegistryValueRecord(
                ts=key.timestamp,
                path=parent,
                key=key.name,
                name=value.name,
                value=value.value,
                data_type=value.type,
                source=path,
                _target=self.target,
            )

        for subkey in key.subkeys():
            n_parent = f"{parent}\\{subkey.name}"
            yield from self.walk(subkey, n_parent, path)



from __future__ import annotations

import io
import zipfile

import pytest

from dissect.target.exceptions import (
    IsADirectoryError,
    NotADirectoryError,
    NotASymlinkError,
)
from dissect.target.filesystems.zip import ZipFilesystem, ZipFilesystemEntry


def _mkdir(zf: zipfile.ZipFile, name: str) -> None:
    # There's no easy way to make a directory with Python zipfile, so copy what zipfile.py does
    zinfo = zipfile.ZipInfo(name)
    zinfo.compress_size = 0
    zinfo.CRC = 0
    zinfo.external_attr = ((0o40000 | 511) & 0xFFFF) << 16
    zinfo.file_size = 0
    zinfo.external_attr |= 0x10

    with zf._lock:
        if zf._seekable:
            zf.fp.seek(zf.start_dir)
        zinfo.header_offset = zf.fp.tell()  # Start of header bytes
        if zinfo.compress_type == zipfile.ZIP_LZMA:
            # Compressed data includes an end-of-stream (EOS) marker
            zinfo.flag_bits |= zipfile._MASK_COMPRESS_OPTION_1

        zf._writecheck(zinfo)
        zf._didModify = True

        zf.filelist.append(zinfo)
        zf.NameToInfo[zinfo.filename] = zinfo
        zf.fp.write(zinfo.FileHeader(False))
        zf.start_dir = zf.fp.tell()


def _create_zip(prefix: str = "", zip_dir: bool = True) -> io.BytesIO:
    buf = io.BytesIO()
    zf = zipfile.ZipFile(buf, "w")

    if prefix and zip_dir:
        cur = []
        for p in prefix.strip("/").split("/"):
            cur.append(p)
            _mkdir(zf, "/".join(cur))

    zf.writestr(zipfile.ZipInfo(f"{prefix}file_1", (1980, 0, 0, 0, 0, 0)), "file 1 contents")
    zf.writestr(zipfile.ZipInfo(f"{prefix}file_2", (2107, 1, 1, 0, 0, 0)), "file 2 contents")
    zf.writestr(zipfile.ZipInfo(f"{prefix}file_3", (1980, 1, 0, 0, 0, 0)), "file 3 contents")
    zf.writestr(zipfile.ZipInfo(f"{prefix}file_4", (2107, 13, 1, 0, 0, 0)), "file 4 contents")
    zf.writestr(zipfile.ZipInfo(f"{prefix}file_5", (2025, 9, 8, 10, 39, 40)), "file 5 contents")

    if zip_dir:
        _mkdir(zf, f"{prefix}dir/")

    for i in range(100):
        zf.writestr(f"{prefix}dir/{i}", f"contents {i}")

    symlink = zipfile.ZipInfo(f"{prefix}symlink_dir")
    symlink.external_attr = 0o120777 << 16
    zf.writestr(symlink, "dir/")

    symlink = zipfile.ZipInfo(f"{prefix}symlink_file")
    symlink.external_attr = 0o120777 << 16
    zf.writestr(symlink, "file_1")

    zf.close()
    buf.seek(0)
    return buf


@pytest.fixture
def zip_simple() -> io.BytesIO:
    return _create_zip()


@pytest.fixture
def zip_base() -> io.BytesIO:
    return _create_zip("base/")


@pytest.fixture
def zip_relative() -> io.BytesIO:
    return _create_zip("./", False)


@pytest.fixture
def zip_relative_dir() -> io.BytesIO:
    return _create_zip("./")


@pytest.fixture
def zip_virtual_dir() -> io.BytesIO:
    return _create_zip("", False)


@pytest.mark.parametrize(
    ("obj", "base"),
    [
        ("zip_simple", ""),
        ("zip_base", "base/"),
        ("zip_relative", ""),
        ("zip_relative_dir", ""),
        ("zip_virtual_dir", ""),
    ],
)
def test_filesystems_zip(obj: str, base: str, request: pytest.FixtureRequest) -> None:
    fh = request.getfixturevalue(obj)

    assert ZipFilesystem.detect(fh)

    fs = ZipFilesystem(fh, base)
    assert isinstance(fs, ZipFilesystem)
    assert len(fs.listdir("/")) == 8

    assert fs.get("./file_1").open().read() == b"file 1 contents"
    assert fs.get("./file_2").open().read() == b"file 2 contents"
    assert fs.get("./file_3").open().read() == b"file 3 contents"
    assert fs.get("./file_1").lstat().st_mtime_ns == 315532800000000000
    assert fs.get("./file_2").lstat().st_mtime_ns == 4323283200000000000
    assert fs.get("./file_3").lstat().st_mtime_ns == 315532800000000000
    assert fs.get("./file_4").lstat().st_mtime_ns == 4354819199000000000
    assert fs.get("./file_5").lstat().st_mtime_ns == 1757327980000000000
    assert fs.get("./symlink_file").open().read() == b"file 1 contents"
    assert len(list(fs.glob("./dir/*"))) == 100
    assert len(list(fs.glob("./symlink_dir/*"))) == 100

    zfile = fs.get("./file_1")
    zdir = fs.get("./dir")
    zsymd = fs.get("./symlink_dir")
    zsymf = fs.get("./symlink_file")

    assert zfile.is_file()
    assert not zfile.is_dir()
    assert not zfile.is_symlink()

    with pytest.raises(NotADirectoryError):
        list(zfile.iterdir())

    with pytest.raises(NotADirectoryError):
        next(zfile.scandir())

    with pytest.raises(NotASymlinkError):
        zfile.readlink()

    assert zdir.is_dir()
    assert not zdir.is_file()
    assert not zdir.is_symlink()
    assert len(list(zdir.iterdir())) == 100
    assert len(list(zdir.scandir())) == 100

    with pytest.raises(IsADirectoryError):
        zdir.open()

    assert zsymd.is_dir()
    assert not zsymd.is_file()
    assert zsymd.is_symlink()
    assert zsymd.readlink() == "dir/"

    assert not zsymf.is_dir()
    assert zsymf.is_file()
    assert zsymf.is_symlink()
    assert zsymf.readlink() == "file_1"

    file1 = zdir.get("1")
    assert file1.is_file()
    assert not file1.is_dir()
    assert not file1.is_symlink()
    assert file1.open().read() == b"contents 1"

    assert file1.stat() == zsymd.get("1").stat()

    assert zfile.stat().st_mode == 0o100600
    assert zfile.stat(follow_symlinks=False) == zfile.lstat()

    if isinstance(zdir, ZipFilesystemEntry):
        assert zdir.stat().st_mode == 0o40777



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.filesystems.squashfs import SquashFSFilesystem
from dissect.target.helpers import configutil
from dissect.target.helpers.fsutil import TargetPath
from dissect.target.plugin import Plugin, alias, export
from dissect.target.plugins.os.unix.applications import UnixApplicationRecord
from dissect.target.target import Target

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.helpers.fsutil import TargetPath
    from dissect.target.target import Target


class SnapPlugin(Plugin):
    """Canonical Linux Snapcraft plugin."""

    PATHS = ("/var/lib/snapd/snaps",)

    def __init__(self, target: Target):
        super().__init__(target)
        self.installs = list(self._find_installs())

    def check_compatible(self) -> None:
        if not configutil.HAS_YAML:
            raise UnsupportedPluginError("Missing required dependency ruamel.yaml")

        if not self.installs:
            raise UnsupportedPluginError("No snapd install folder(s) found")

    def _find_installs(self) -> Iterator[TargetPath]:
        for str_path in self.PATHS:
            if (path := self.target.fs.path(str_path)).exists():
                yield path

    @export(record=UnixApplicationRecord)
    @alias("snaps")
    def snap(self) -> Iterator[UnixApplicationRecord]:
        """Yields installed Canonical Linux Snapcraft (snaps) applications on the target system.

        Reads information from installed SquashFS ``*.snap`` files found in ``/var/lib/snapd/snaps``.
        Logs of the ``snapd`` daemon can be parsed using the ``journal`` or ``syslog`` plugins.

        References:
            - https://github.com/canonical/snapcraft
            - https://en.wikipedia.org/wiki/Snap_(software)

        Yields ``UnixApplicationRecord`` records with the following fields:

        .. code-block:: text

            ts_modified  (datetime): timestamp when the installation was modified
            name         (string):   name of the application
            version      (string):   version of the application
            path         (string):   path to the application snap file
        """

        for install_path in self.installs:
            for snap in install_path.glob("*.snap"):
                try:
                    squashfs = SquashFSFilesystem(snap.open())

                except (ValueError, NotImplementedError) as e:
                    self.target.log.warning("Unable to open snap file %s", snap)
                    self.target.log.debug("", exc_info=e)
                    continue

                if not (meta := squashfs.path("meta/snap.yaml")).exists():
                    self.target.log.warning("Snap %s has no meta/snap.yaml file")
                    continue

                meta_data = configutil.parse(meta, hint="yaml")

                yield UnixApplicationRecord(
                    ts_modified=meta.lstat().st_mtime,
                    name=meta_data.get("name"),
                    version=meta_data.get("version"),
                    path=snap,
                    _target=self.target,
                )



from __future__ import annotations

from io import BytesIO
from typing import BinaryIO
from unittest.mock import patch

from dissect.target import volume
from dissect.target.volumes import disk


def test_reset_file_position() -> None:
    fh = BytesIO(b"\x00" * 8192)
    fh.seek(512)

    class MockVolumeSystem(volume.VolumeSystem):
        def __init__(self, fh: BinaryIO):
            assert fh.tell() == 0
            fh.seek(1024)
            self.success = True

        @staticmethod
        def _detect(fh: BinaryIO) -> bool:
            assert fh.tell() == 0
            fh.seek(256)
            return True

    with patch.object(disk, "DissectVolumeSystem", MockVolumeSystem):
        assert MockVolumeSystem.detect(fh)
        assert fh.tell() == 512

        opened_vs = volume.open(fh)
        assert isinstance(opened_vs, MockVolumeSystem)
        assert opened_vs.success
        assert fh.tell() == 512



from __future__ import annotations

import io
import logging
import typing

from dissect.target import Target
from dissect.target.filesystem import VirtualFilesystem

if typing.TYPE_CHECKING:
    from pathlib import Path

    import pytest


def test_direct_overlap_warning(
    tmp_path: Path, caplog: pytest.LogCaptureFixture, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Assert direct raise warning in case sensitive mode if some files overlap
    We must uncompress files in a temporary directory as having two files with same name
    would cause issue with git on case insensitive fs.
    """
    source_vfs = VirtualFilesystem(case_sensitive=True)
    source_vfs.map_file_fh("file.txt", io.BytesIO(b"a"))
    source_vfs.map_file_fh("File.txt", io.BytesIO(b"b"))
    with caplog.at_level(logging.WARNING):
        _ = Target.open_direct([source_vfs.path("/")], case_sensitive=True)
        assert (
            "Direct mode used in case insensitive mode, but this will cause files overlap, "
            "consider using --direct-sensitive" not in caplog.text
        )
    with caplog.at_level(logging.WARNING):
        _ = Target.open_direct([source_vfs.path("/")], case_sensitive=False)
        assert (
            "Direct mode used in case insensitive mode, but this will cause files overlap, "
            "consider using --direct-sensitive" in caplog.text
        )
    with caplog.at_level(logging.WARNING):
        _ = Target.open_direct([source_vfs.path("/file.txt"), source_vfs.path("/File.txt")], case_sensitive=False)
        assert (
            "Direct mode used in case insensitive mode, but this will cause files overlap, "
            "consider using --direct-sensitive" in caplog.text
        )



from __future__ import annotations

import io
import tarfile

import pytest

from dissect.target.exceptions import IsADirectoryError, NotADirectoryError
from dissect.target.filesystems.tar import TarFilesystem, TarFilesystemEntry


def _mkdir(tf: tarfile.TarFile, name: str) -> None:
    info = tf.tarinfo()
    info.name = name
    info.mode = 0o40777
    info.uid = 0
    info.gid = 0
    info.size = 0
    info.mtime = 3384460800
    info.type = tarfile.DIRTYPE
    info.linkname = ""
    tf.addfile(info)


def _mkfile(tf: tarfile.TarFile, name: str, content: bytes) -> None:
    info = tf.tarinfo()
    info.name = name
    info.mode = 0o100777
    info.uid = 0
    info.gid = 0
    info.size = len(content)
    info.mtime = 3384460800
    info.type = tarfile.REGTYPE
    info.linkname = ""
    tf.addfile(info, io.BytesIO(content))


def _mksym(tf: tarfile.TarFile, name: str, dest: str) -> None:
    info = tf.tarinfo()
    info.name = name
    info.mode = 0o120777
    info.uid = 0
    info.gid = 0
    info.size = 0
    info.mtime = 3384460800
    info.type = tarfile.SYMTYPE
    info.linkname = dest
    tf.addfile(info)


def _create_tar(prefix: str = "", tar_dir: bool = True, tar_sym: bool = False) -> io.BytesIO:
    buf = io.BytesIO()
    tf = tarfile.TarFile(fileobj=buf, mode="w")

    if prefix and tar_dir:
        cur = []
        for p in prefix.strip("/").split("/"):
            cur.append(p)
            _mkdir(tf, "/".join(cur))

    _mkfile(tf, f"{prefix}file_1", b"file 1 contents")
    _mkfile(tf, f"{prefix}file_2", b"file 2 contents")

    if tar_dir:
        _mkdir(tf, f"{prefix}dir/")

    for i in range(100):
        _mkfile(tf, f"{prefix}dir/{i}", f"contents {i}".encode())

    if tar_sym:
        _mksym(tf, f"{prefix}sym_1", f"{prefix}file_1")
        _mksym(tf, f"{prefix}sym_2", f"{prefix}dir")

    tf.close()
    buf.seek(0)
    return buf


@pytest.fixture
def tar_simple() -> io.BytesIO:
    return _create_tar()


@pytest.fixture
def tar_base() -> io.BytesIO:
    return _create_tar("base/")


@pytest.fixture
def tar_relative() -> io.BytesIO:
    return _create_tar("./", False)


@pytest.fixture
def tar_relative_dir() -> io.BytesIO:
    return _create_tar("./")


@pytest.fixture
def tar_virtual_dir() -> io.BytesIO:
    return _create_tar("", False)


@pytest.fixture
def tar_symlink() -> io.BytesIO:
    return _create_tar("", tar_sym=True)


@pytest.mark.parametrize(
    ("obj", "base"),
    [
        ("tar_simple", ""),
        ("tar_base", "base/"),
        ("tar_relative", ""),
        ("tar_relative_dir", ""),
        ("tar_virtual_dir", ""),
        ("tar_symlink", ""),
    ],
)
def test_filesystems_tar(obj: str, base: str, request: pytest.FixtureRequest) -> None:
    fh = request.getfixturevalue(obj)

    assert TarFilesystem.detect(fh)

    fs = TarFilesystem(fh, base)
    assert isinstance(fs, TarFilesystem)

    assert len(fs.listdir("/")) == (5 if fs.lexists("sym_1") else 3)

    assert fs.get("./file_1").open().read() == b"file 1 contents"
    assert fs.get("./file_2").open().read() == b"file 2 contents"
    assert len(list(fs.glob("./dir/*"))) == 100

    tfile = fs.get("./file_1")
    tdir = fs.get("./dir")

    assert tfile.is_file()
    assert not tfile.is_dir()
    assert not tfile.is_symlink()

    with pytest.raises(NotADirectoryError):
        tfile.listdir()

    assert tdir.is_dir()
    assert not tdir.is_file()
    assert not tdir.is_symlink()

    with pytest.raises(IsADirectoryError):
        tdir.open()

    file1 = tdir.get("1")
    assert file1.is_file()
    assert not file1.is_dir()
    assert not file1.is_symlink()
    assert file1.open().read() == b"contents 1"

    assert tfile.stat().st_mode == 0o100777

    if isinstance(tdir, TarFilesystemEntry):
        assert tdir.stat().st_mode == 0o40777

    if fs.lexists("sym_1"):
        tsymf = fs.get("sym_1")

        assert tsymf.is_file()
        assert not tsymf.is_file(follow_symlinks=False)
        assert not tsymf.is_dir()
        assert tsymf.is_symlink()

        assert tsymf.lstat().st_mode == 0o120777
        assert tsymf.stat().st_mode == 0o100777
        assert tsymf.stat(follow_symlinks=False) == tsymf.lstat()

        with pytest.raises(NotADirectoryError):
            tsymf.listdir()

        assert tsymf.open().read() == b"file 1 contents"

    if fs.lexists("sym_2"):
        tsymd = fs.get("sym_2")

        assert not tsymd.is_file()
        assert tsymd.is_dir()
        assert not tsymd.is_dir(follow_symlinks=False)
        assert tsymd.is_symlink()

        assert tsymd.lstat().st_mode == 0o120777
        assert tsymd.stat().st_mode == 0o40777
        assert tsymd.stat(follow_symlinks=False) == tsymd.lstat()

        with pytest.raises(IsADirectoryError):
            tsymd.open()

        assert len(list(tsymd.listdir())) == 100



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.plugins.apps.vpn.wireguard import WireGuardPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_wireguard_plugin_global_log(target_unix_users: Target, fs_unix: VirtualFilesystem) -> None:
    wireguard_config_file = absolute_path("_data/plugins/apps/vpn/wireguard/wg0.conf")
    fs_unix.map_file("/etc/wireguard/wg0.conf", wireguard_config_file)

    target_unix_users.add_plugin(WireGuardPlugin)
    records = list(target_unix_users.wireguard.config())
    assert len(records) == 6

    # Interface
    record = records[0]
    assert record.name == "wg0"
    assert str(record.address) == "10.13.37.1"
    assert record.private_key == "UHJpdmF0ZUtleQ=="
    assert record.listen_port == 12345
    assert record.source == "/etc/wireguard/wg0.conf"
    assert record.dns == []
    assert record.table == "1337"

    # Peer
    record = records[1]
    assert record.name is None
    assert [str(addr) for addr in record.allowed_ips] == ["10.13.37.2/32", "::/0"]
    assert record.public_key == "UHVibGljS2V5MQ=="
    assert record.source == "/etc/wireguard/wg0.conf"

    # Peer
    record = records[2]
    assert record.name is None
    assert [str(addr) for addr in record.allowed_ips] == ["10.13.37.3/32", "::/0"]
    assert record.public_key == "UHVibGljS2V5Mg=="
    assert record.source == "/etc/wireguard/wg0.conf"

    # Peer
    record = records[3]
    assert record.name is None
    assert record.allowed_ips == ["10.13.37.4"]
    assert record.public_key == "UHVibGljS2V5Mg=="
    assert record.source == "/etc/wireguard/wg0.conf"

    # Interface
    record = records[4]
    assert record.name == "wg0"
    assert record.address == "10.13.37.5"
    assert record.private_key == "UHJpdmF0ZUtleTI=="
    assert not record.listen_port
    assert record.source == "/etc/wireguard/wg0.conf"
    assert record.dns == ["10.13.37.1", "1.1.1.1"]
    assert record.table == "off"

    # Peer
    record = records[5]
    assert record.name is None
    assert record.allowed_ips == ["0.0.0.0/0"]
    assert record.public_key == "TWFkZVlvdUxvb2s="
    assert record.source == "/etc/wireguard/wg0.conf"



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.util import ts

from dissect.target.plugin import Plugin, export
from dissect.target.plugins.os.unix.generic import calculate_last_activity

if TYPE_CHECKING:
    from datetime import datetime


class GenericPlugin(Plugin):
    """Generic plugin for iOS targets."""

    def check_compatible(self) -> None:
        pass

    @export(property=True)
    def activity(self) -> datetime | None:
        """Return last seen activity based on filesystem timestamps."""

        return calculate_last_activity(self.target.fs.path("/private/var/log"), recursive=True)

    @export(property=True)
    def install_date(self) -> datetime | None:
        """Return the likely install date of the operating system."""

        # prng.seed seems to be created when iOS is first installed and not touched afterwards,
        # this is an educated guess without any further research.
        for path in ["/private/var/db/prng.seed"]:
            if (file := self.target.fs.path(path)).exists():
                return ts.from_unix(file.lstat().st_mtime)
        return None



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from dissect.target.loader import open as loader_open
from dissect.target.loaders.vma import VmaLoader
from dissect.target.target import Target
from tests._utils import absolute_path

if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
def test_target_open(opener: Callable[[str | Path], Target]) -> None:
    """Test that we correctly use ``VmaLoader`` when opening a ``Target``."""
    path = absolute_path("_data/loaders/vma/vzdump-qemu-6969-2025_08_01-15_24_07.vma")

    target = opener(path)
    assert isinstance(target._loader, VmaLoader)
    assert target.path == path


def test_loader() -> None:
    """Test that ``VmaLoader`` correctly loads a VMA file and its disks."""
    path = absolute_path("_data/loaders/vma/vzdump-qemu-6969-2025_08_01-15_24_07.vma")

    loader = loader_open(path)
    assert isinstance(loader, VmaLoader)

    t = Target()
    loader.map(t)

    assert len(t.disks) == 1
    assert t.disks[0].size == 1073741824



from __future__ import annotations

import textwrap
from io import BytesIO
from typing import TYPE_CHECKING

from dissect.util.ts import from_unix

from dissect.target.plugins.os.unix.log.audit import AuditPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_audit_plugin(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/os/unix/log/audit/audit.log")
    fs_unix.map_file("var/log/audit/audit.log", data_file)

    target_unix.add_plugin(AuditPlugin)
    results = list(target_unix.audit())

    assert len(results) == 4

    result = results[0]
    assert result.audit_type == "SYSCALL"
    assert result.ts == from_unix(1364481363.243)
    assert result.audit_id == 24287
    assert (
        result.message
        == 'arch=c000003e syscall=2 success=no exit=-13 a0=7fffd19c5592 a1=0 a2=7fffd19c4b50 a3=a items=1 ppid=2686 pid=3538 auid=1000 uid=1000 gid=1000 euid=1000 suid=1000 fsuid=1000 egid=1000 sgid=1000 fsgid=1000 tty=pts0 ses=1 comm="cat" exe="/bin/cat" subj=unconfined_u:unconfined_r:unconfined_t:s0-s0:c0.c1023 key="sshd_config"'  # noqa E501
    )

    result = results[1]
    assert result.audit_type == "CWD"
    assert result.ts == from_unix(1364481363.243)
    assert result.audit_id == 24287
    assert result.message == 'cwd="/home/shadowman"'


def test_audit_plugin_config(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    config = """
    log_file = /foo/bar/audit/audit.log
    # log_file=/tmp/disabled/audit/audit.log
    """
    fs_unix.map_file_fh("etc/audit/auditd.conf", BytesIO(textwrap.dedent(config).encode()))
    fs_unix.map_file_fh("tmp/disabled/audit/audit.log", BytesIO(b"Foo"))
    fs_unix.map_file_fh("foo/bar/audit/audit.log", BytesIO(b"Foo"))

    log_paths = AuditPlugin(target_unix).get_log_paths()
    assert len(log_paths) == 2
    assert str(log_paths[0]) == "/foo/bar/audit/audit.log"
    assert str(log_paths[1]) == "/tmp/disabled/audit/audit.log"
